import io
import operator
import os
import sys
import unittest
from typing import Callable

import pyrtl
from pyrtl.rtllib import testingutils as utils
from tests.test_transform import NetWireNumTestCases


class TestSynthesis(unittest.TestCase):
    def setUp(self):
        pyrtl.reset_working_block()
        self.bitwidth = 3
        self.r = pyrtl.Register(bitwidth=self.bitwidth)
        self.output = pyrtl.Output(bitwidth=self.bitwidth, name="r")
        self.output <<= self.r

    def check_trace(self, correct_string):
        pyrtl.synthesize()
        sim = pyrtl.Simulation()
        sim.step_multiple(nsteps=8)
        output = io.StringIO()
        sim.tracer.print_trace(output, compact=True)
        self.assertEqual(output.getvalue(), correct_string)

    def test_not_simulation(self):
        self.r.next <<= ~self.r
        self.check_trace("r 07070707\n")

    def test_and_simulation(self):
        self.r.next <<= (~self.r) & pyrtl.Const(6, bitwidth=self.bitwidth)
        self.check_trace("r 06060606\n")

    def test_or_simulation(self):
        result = self.r | pyrtl.Const(4, bitwidth=self.bitwidth)
        self.r.next <<= result
        self.assertEqual(len(result), self.bitwidth)
        self.check_trace("r 04444444\n")

    def test_xor_simulation(self):
        self.r.next <<= self.r ^ pyrtl.Const(4, bitwidth=self.bitwidth)
        self.check_trace("r 04040404\n")

    def test_plus_simulation(self):
        self.r.next <<= self.r + pyrtl.Const(2, bitwidth=self.bitwidth)
        self.check_trace("r 02460246\n")

    def test_minus_simulation(self):
        self.r.next <<= self.r - pyrtl.Const(1, bitwidth=self.bitwidth)
        self.check_trace("r 07654321\n")

    def test_minus_simulation2(self):
        self.r.next <<= self.r - pyrtl.Const(3, bitwidth=self.bitwidth)
        self.check_trace("r 05274163\n")

    def test_const_nobitwidth_simulation(self):
        self.r.next <<= self.r - pyrtl.Const(1)
        self.check_trace("r 07654321\n")

    def test_mux_simulation(self):
        self.r.next <<= pyrtl.mux(self.r, 4, 3, 1, 7, 2, 6, 0, 5)
        self.check_trace("r 04213756\n")

    def test_synthesize_regs_mapped_correctly(self):
        r2 = pyrtl.Register(5)
        self.r.next <<= ~self.r
        r2.next <<= self.r + 1
        synth_block = pyrtl.synthesize()
        self.assertEqual(len(synth_block.reg_map), 2)
        self.assertEqual(len(synth_block.reg_map[self.r]), len(self.r))
        self.assertEqual(len(synth_block.reg_map[r2]), len(r2))


class TestIOInterfaceSynthesis(unittest.TestCase):
    def setUp(self):
        pyrtl.reset_working_block()
        a, b = pyrtl.input_list("a/4 b/4")
        o = pyrtl.Output(5, "o")
        o <<= a + b

    def check_merged_names(self, block):
        inputs = block.wirevector_subset(pyrtl.Input)
        outputs = block.wirevector_subset(pyrtl.Output)
        self.assertEqual({w.name for w in inputs}, {"a", "b"})
        self.assertEqual({w.name for w in outputs}, {"o"})

    def check_unmerged_names(self, block):
        inputs = block.wirevector_subset(pyrtl.Input)
        outputs = block.wirevector_subset(pyrtl.Output)
        self.assertEqual(
            {w.name for w in inputs},
            {"a[0]", "a[1]", "a[2]", "a[3]", "b[0]", "b[1]", "b[2]", "b[3]"},
        )
        self.assertEqual(
            {w.name for w in outputs}, {"o[0]", "o[1]", "o[2]", "o[3]", "o[4]"}
        )

    def test_synthesize_merged_io_names_correct(self):
        pyrtl.synthesize()
        self.check_merged_names(pyrtl.working_block())

    def test_synthesize_merged_io_mapped_correctly(self):
        old_io = pyrtl.working_block().wirevector_subset((pyrtl.Input, pyrtl.Output))
        pyrtl.synthesize()
        new_io = pyrtl.working_block().wirevector_subset((pyrtl.Input, pyrtl.Output))
        for oi in old_io:
            io_list = pyrtl.working_block().io_map[oi]
            self.assertEqual(len(io_list), 1)
            for ni in new_io:
                if oi.name == ni.name:
                    self.assertEqual(io_list, [ni])

    def test_synthesize_merged_io_simulates_correctly(self):
        pyrtl.synthesize()
        sim = pyrtl.Simulation()
        sim.step_multiple(
            {
                "a": [4, 6, 2, 3],
                "b": [2, 9, 11, 4],
            }
        )
        output = io.StringIO()
        sim.tracer.print_trace(output, compact=True)
        self.assertEqual(output.getvalue(), "a 4623\nb 29114\no 615137\n")

    def test_synthesize_unmerged_io_names_correct(self):
        pyrtl.synthesize(merge_io_vectors=False)
        self.check_unmerged_names(pyrtl.working_block())

    def test_synthesize_unmerged_io_mapped_correctly(self):
        old_io = pyrtl.working_block().wirevector_subset((pyrtl.Input, pyrtl.Output))
        pyrtl.synthesize(merge_io_vectors=False)
        new_io = pyrtl.working_block().wirevector_subset((pyrtl.Input, pyrtl.Output))
        for oi in old_io:
            io_list = [w.name for w in pyrtl.working_block().io_map[oi]]
            self.assertEqual(len(io_list), len(oi))
            for ni in new_io:
                if ni.name.startswith(oi.name):
                    # Dev note: comparing names because comparing wires (e.g. list/set
                    # inclusion) creates an '=' net, which is definitely not what we
                    # want here.
                    self.assertIn(ni.name, io_list)

    def test_synthesize_unmerged_io_simulates_correctly(self):
        pyrtl.synthesize(merge_io_vectors=False)
        sim = pyrtl.Simulation()
        for a, b in [(4, 2), (6, 9), (2, 11), (3, 4)]:
            args = {}
            for ix in range(4):
                args["a[" + str(ix) + "]"] = (a >> ix) & 1
                args["b[" + str(ix) + "]"] = (b >> ix) & 1
            sim.step(args)
            expected = a + b
            for ix in range(5):
                out = sim.inspect("o[" + str(ix) + "]")
                self.assertEqual(out, (expected >> ix) & 1)

    def test_synthesize_does_not_update_working_block(self):
        synth_block = pyrtl.synthesize(
            update_working_block=False, merge_io_vectors=False
        )
        self.check_merged_names(pyrtl.working_block())
        self.check_unmerged_names(synth_block)


class TestMultiplierSynthesis(unittest.TestCase):
    def setUp(self):
        pyrtl.reset_working_block()
        self.output = pyrtl.Output(name="r")

    def test_single_mul(self):
        ina, inb = pyrtl.Input(bitwidth=4, name="a"), pyrtl.Input(bitwidth=4, name="b")
        self.output <<= ina * inb
        pyrtl.synthesize()
        sim = pyrtl.Simulation()
        for a in range(16):
            for b in range(16):
                sim.step({"a": a, "b": b})
        result = sim.tracer.trace["r"]
        self.assertEqual(result, [a * b for a in range(16) for b in range(16)])

    def test_chained_mul(self):
        ina, inb, inc = (
            pyrtl.Input(bitwidth=2, name="a"),
            pyrtl.Input(bitwidth=2, name="b"),
            pyrtl.Input(bitwidth=2, name="c"),
        )
        self.output <<= ina * inb * inc
        pyrtl.synthesize()
        sim = pyrtl.Simulation()
        for a in range(4):
            for b in range(4):
                for c in range(4):
                    sim.step({"a": a, "b": b, "c": c})
        result = sim.tracer.trace["r"]
        self.assertEqual(
            result, [a * b * c for a in range(4) for b in range(4) for c in range(4)]
        )

    def test_singlebit_mul(self):
        ina, inb = pyrtl.Input(bitwidth=1, name="a"), pyrtl.Input(bitwidth=3, name="b")
        self.output <<= ina * inb
        pyrtl.synthesize()
        sim = pyrtl.Simulation()
        for a in range(2):
            for b in range(8):
                sim.step({"a": a, "b": b})
        result = sim.tracer.trace["r"]
        self.assertEqual(result, [a * b for a in range(2) for b in range(8)])


class TestComparisonSynthesis(unittest.TestCase):
    def setUp(self):
        pyrtl.reset_working_block()
        self.output = pyrtl.Output(name="r")

    def check_op(self, op):
        ina, inb = pyrtl.Input(bitwidth=4, name="a"), pyrtl.Input(bitwidth=4, name="b")
        self.output <<= op(ina, inb)
        pyrtl.synthesize()
        sim = pyrtl.Simulation()
        for a in range(16):
            for b in range(16):
                sim.step({"a": a, "b": b})
        result = sim.tracer.trace["r"]
        self.assertEqual(result, [op(a, b) for a in range(16) for b in range(16)])

    def test_eq(self):
        self.check_op(lambda x, y: x == y)

    def test_lt(self):
        self.check_op(lambda x, y: x < y)

    def test_gt(self):
        self.check_op(lambda x, y: x > y)


class TestOptimization(NetWireNumTestCases):
    def test_wire_net_removal_1(self):
        inwire = pyrtl.Input(bitwidth=3)
        tempwire = pyrtl.WireVector()
        outwire = pyrtl.Output()
        tempwire <<= inwire
        outwire <<= tempwire
        pyrtl.synthesize()
        pyrtl.optimize()
        block = pyrtl.working_block()
        # should remove the middle wire but keep the input
        self.assert_num_net(5, block)
        self.assert_num_wires(6, block)

    def test_wire_net_removal_2(self):
        inwire = pyrtl.Input(bitwidth=3)
        tempwire = pyrtl.WireVector()
        tempwire2 = pyrtl.WireVector()
        outwire = pyrtl.Output()
        tempwire <<= inwire
        tempwire2 <<= tempwire
        outwire <<= tempwire
        pyrtl.synthesize()
        pyrtl.optimize()
        # should remove the middle wires but keep the input
        block = pyrtl.working_block()
        self.assert_num_net(5, block)
        self.assert_num_wires(6, block)

    def test_slice_net_removal_1(self):
        constwire = pyrtl.Const(1, 1)
        inwire = pyrtl.Input(bitwidth=1)
        outwire = pyrtl.Output()
        outwire <<= constwire ^ inwire
        pyrtl.optimize()
        block = pyrtl.working_block()
        self.num_net_of_type("s", 0, block)
        self.num_net_of_type("~", 1, block)

    def test_slice_net_removal_2(self):
        inwire = pyrtl.Input(bitwidth=3)
        outwire = pyrtl.Output()
        tempwire = inwire[0:3]
        outwire <<= tempwire[0:3]
        pyrtl.optimize()
        block = pyrtl.working_block()
        self.num_net_of_type("s", 0, block)
        self.num_net_of_type("w", 1, block)

    def test_slice_net_removal_3(self):
        inwire = pyrtl.Input(bitwidth=3)
        outwire = pyrtl.Output()
        tempwire = inwire[0:2]
        outwire <<= tempwire[0:2]
        pyrtl.optimize()
        # Removes one of the slices, which does nothing.
        block = pyrtl.working_block()
        self.num_net_of_type("s", 1, block)
        self.num_net_of_type("w", 1, block)

    def test_slice_net_removal_4(self):
        inwire = pyrtl.Input(bitwidth=4)
        outwire1 = pyrtl.Output()
        outwire2 = pyrtl.Output()
        outwire1 <<= inwire[0:4]
        outwire2 <<= inwire[0:3]
        pyrtl.optimize()
        # Removes just the outwire1 slice, which does nothing.
        block = pyrtl.working_block()
        self.num_net_of_type("s", 1, block)
        self.num_net_of_type("w", 2, block)

    def test_remove_double_inverts_1_invert(self):
        inwire = pyrtl.Input(bitwidth=1)
        outwire = pyrtl.Output(bitwidth=1)
        outwire <<= ~inwire
        pyrtl.optimize()
        block = pyrtl.working_block()
        self.assert_num_net(2, block)
        self.assert_num_wires(3, block)

    def test_remove_double_inverts_3_inverts(self):
        inwire = pyrtl.Input(bitwidth=1)
        outwire = pyrtl.Output(bitwidth=1)
        outwire <<= ~(~(~inwire))
        pyrtl.optimize()
        block = pyrtl.working_block()
        self.assert_num_net(2, block)
        self.assert_num_wires(3, block)

    def test_remove_double_inverts_5_inverts(self):
        inwire = pyrtl.Input(bitwidth=1)
        outwire = pyrtl.Output(bitwidth=1)
        outwire <<= ~(~(~(~(~inwire))))
        pyrtl.optimize()
        block = pyrtl.working_block()
        self.assert_num_net(2, block)
        self.assert_num_wires(3, block)

    def test_remove_double_inverts_2_inverts(self):
        inwire = pyrtl.Input(bitwidth=1)
        outwire = pyrtl.Output(bitwidth=1)
        outwire <<= ~(~inwire)
        pyrtl.optimize()
        block = pyrtl.working_block()
        self.assert_num_net(1, block)
        self.assert_num_wires(2, block)

    def test_remove_double_inverts_4_inverts(self):
        inwire = pyrtl.Input(bitwidth=1)
        outwire = pyrtl.Output(bitwidth=1)
        outwire <<= ~(~(~(~inwire)))
        pyrtl.optimize()
        block = pyrtl.working_block()
        self.assert_num_net(1, block)
        self.assert_num_wires(2, block)

    def test_remove_double_inverts_6_inverts(self):
        inwire = pyrtl.Input(bitwidth=1)
        outwire = pyrtl.Output(bitwidth=1)
        outwire <<= ~(~(~(~(~(~inwire)))))
        pyrtl.optimize()
        block = pyrtl.working_block()
        self.assert_num_net(1, block)
        self.assert_num_wires(2, block)

    def test_dont_remove_double_inverts_another_user(self):
        inwire = pyrtl.Input(bitwidth=1)
        outwire = pyrtl.Output(bitwidth=1)
        outwire2 = pyrtl.Output(bitwidth=1)
        tempwire = pyrtl.WireVector()
        tempwire <<= ~inwire
        outwire <<= ~tempwire
        outwire2 <<= tempwire
        pyrtl.optimize()
        block = pyrtl.working_block()
        self.assert_num_net(4, block)
        self.assert_num_wires(5, block)

    def test_multiple_double_invert_chains(self):
        # _remove_double_inverts removes double inverts by chains, so it is useful to
        # make sure it can remove double inverts from multiple chains
        inwire = pyrtl.Input(bitwidth=1)
        outwire = pyrtl.Output(bitwidth=1)
        outwire2 = pyrtl.Output(bitwidth=1)
        outwire <<= ~(~inwire)
        outwire2 <<= ~(~(~(~(inwire))))
        pyrtl.optimize()
        block = pyrtl.working_block()
        self.assert_num_net(2, block)
        self.assert_num_wires(3, block)


class TestConstFolding(NetWireNumTestCases):
    def setUp(self):
        pyrtl.reset_working_block()
        # Redirect stdout because we don't care to see the specific messages about wires
        # being deemed useless by optimization sent to stdout.
        sys.stdout = open(os.devnull, "w")  # noqa: SIM115

    def tearDown(self):
        sys.stdout.close()
        sys.stdout = sys.__stdout__

    def test_basic_one_var_op_1(self):
        constwire = pyrtl.Const(0, 1)
        outwire = pyrtl.Output()

        outwire <<= ~constwire
        pyrtl.synthesize()
        pyrtl.optimize()

        block = pyrtl.working_block()
        self.num_net_of_type("~", 0, block)
        self.num_net_of_type("w", 1, block)
        self.assertEqual(len(block.logic), 1)
        self.assertEqual(len(block.wirevector_set), 2)
        self.num_wire_of_type(pyrtl.Const, 1, block)

    def test_adv_one_var_op_1(self):
        constwire = pyrtl.Const(0, 1)
        outwire = pyrtl.Output()
        tempwire = pyrtl.WireVector()
        reg = pyrtl.Register(1, "test register")

        tempwire <<= ~constwire
        reg.next <<= tempwire
        outwire <<= reg
        pyrtl.synthesize()
        pyrtl.optimize()

        block = pyrtl.working_block(None)
        self.num_net_of_type("w", 1, block)
        self.assert_num_net(1, block)
        self.assert_num_wires(2, block)
        self.num_wire_of_type(pyrtl.Const, 1, block)
        self.num_wire_of_type(pyrtl.Output, 1, block)

    def test_adv_one_var_op_2(self):
        # this one tests to see that an input wirevector is properly preserved

        inwire = pyrtl.Input(bitwidth=1)
        outwire = pyrtl.Output()
        tempwire = pyrtl.WireVector()
        reg = pyrtl.Register(1, "test register")

        tempwire <<= ~inwire
        reg.next <<= tempwire
        outwire <<= reg
        pyrtl.synthesize()
        pyrtl.optimize()
        # should remove the and block and replace it with a wire net (to separate the
        # input from the output)
        block = pyrtl.working_block(None)

        # Note: the current implementation still sticks a wire net between a register
        # 'nextsetter' wire and the output wire
        self.num_net_of_type("w", 1, block)
        self.assert_num_net(3, block)
        self.assert_num_wires(4, block)
        self.num_wire_of_type(pyrtl.Const, 0, block)
        self.num_wire_of_type(pyrtl.Output, 1, block)

    def test_basic_two_var_op_1(self):
        inwire = pyrtl.Input(bitwidth=1)
        constwire = pyrtl.Const(0, 1)
        outwire = pyrtl.Output()

        outwire <<= inwire & constwire
        pyrtl.synthesize()
        pyrtl.optimize()
        # should remove the and block and replace it with a wire net (to separate the
        # const from the output)
        block = pyrtl.working_block(None)
        self.num_net_of_type("&", 0, block)
        self.num_net_of_type("w", 1, block)
        self.assert_num_net(1, block)
        self.assert_num_wires(3, block)
        self.num_wire_of_type(pyrtl.Const, 1, block)

    def test_basic_two_var_op_2(self):
        inwire = pyrtl.Input(bitwidth=1)
        constwire = pyrtl.Const(0, 1)
        outwire = pyrtl.Output()

        outwire <<= inwire | constwire
        pyrtl.optimize()
        # should remove the or block and replace it with a wire net (to separate the
        # const from the output)
        block = pyrtl.working_block(None)
        self.num_net_of_type("|", 0, block)
        self.num_net_of_type("w", 1, block)
        self.assert_num_net(1, block)
        self.assert_num_wires(2, block)
        self.num_wire_of_type(pyrtl.Const, 0, block)

    def test_basic_two_var_op_3(self):
        constwire = pyrtl.Const(0, 1)
        outwire = pyrtl.Output()

        # playing with edge cases
        outwire <<= constwire ^ constwire
        pyrtl.synthesize()
        pyrtl.optimize()
        # should remove the and block and replace it with a wirevector (to separate the
        # input from the output)
        block = pyrtl.working_block(None)
        self.num_net_of_type("^", 0, block)
        self.num_net_of_type("w", 1, block)
        self.assert_num_net(1, block)
        self.assert_num_wires(2, block)
        self.num_wire_of_type(pyrtl.Const, 1, block)

    def test_two_var_op_produce_not(self):
        constwire = pyrtl.Const(1, 1)
        inwire = pyrtl.Input(bitwidth=1)
        outwire = pyrtl.Output()

        # playing with edge cases
        outwire <<= constwire ^ inwire
        pyrtl.synthesize()
        pyrtl.optimize()
        # should remove the and block and replace it with a wirevector (to separate the
        # input from the output)
        block = pyrtl.working_block(None)
        self.num_net_of_type("~", 1, block)
        self.num_net_of_type("w", 1, block)
        self.num_net_of_type("s", 0, block)
        self.assert_num_net(2, block)
        self.assert_num_wires(3, block)
        self.num_wire_of_type(pyrtl.Const, 0, block)

    def test_two_var_op_correct_wire_prop(self):
        ins = [pyrtl.Input(1) for i in range(3)]
        outwire = pyrtl.Output()
        const_w = pyrtl.Const(0)
        temp1 = ins[1] & ins[2]
        temp2 = temp1 ^ const_w
        temp3 = const_w | temp2

        outwire <<= temp3 & ins[2]

        pyrtl.optimize()
        block = pyrtl.working_block()
        block.sanity_check()  # just in case

        self.num_net_of_type("&", 2)
        self.num_net_of_type("w", 1)
        self.assert_num_net(3)
        self.assert_num_wires(6)
        self.num_wire_of_type(pyrtl.Const, 0)

    def test_two_var_op_correct_not_wire_replacement(self):
        ins = [pyrtl.Input(1) for i in range(3)]
        outwire = pyrtl.Output()
        const_0 = pyrtl.Const(0)
        const_1 = pyrtl.Const(1)
        temp1 = ins[0] & ins[1]
        temp2 = const_0 | temp1
        temp3 = temp2 ^ const_1

        outwire <<= temp3 & ins[2]

        pyrtl.optimize()
        block = pyrtl.working_block()
        block.sanity_check()  # just in case

        self.num_net_of_type("&", 2)
        self.num_net_of_type("~", 1)
        self.num_net_of_type("w", 1)
        self.assert_num_net(4)
        self.assert_num_wires(7)
        self.num_wire_of_type(pyrtl.Const, 0)

    def test_unsynthesized_inversion(self):
        constwire = pyrtl.Const(0b1101, bitwidth=4)
        outwire = pyrtl.Output(name="out")

        outwire <<= ~constwire
        pyrtl.optimize()

        sim = pyrtl.Simulation()
        sim.step()
        self.assertEqual(sim.inspect("out"), 0b0010)


class TestSubexpElimination(NetWireNumTestCases):
    def test_basic_1(self):
        ins = [pyrtl.Input(5) for i in range(2)]
        outs = [pyrtl.Output(5) for i in range(2)]
        outs[0] <<= ins[1] & ins[0]
        outs[1] <<= ins[1] & ins[0]

        pyrtl.common_subexp_elimination()
        self.num_net_of_type("&", 1)
        self.num_net_of_type("w", 2)
        self.assert_num_net(3)
        self.assert_num_wires(5)
        pyrtl.working_block().sanity_check()

    def test_different_arg_order(self):
        ins = [pyrtl.Input(5) for i in range(2)]
        outs = [pyrtl.Output(5) for i in range(2)]
        outs[0] <<= ins[1] & ins[0]
        outs[1] <<= ins[0] & ins[1]

        pyrtl.common_subexp_elimination()
        self.num_net_of_type("&", 1)
        self.num_net_of_type("w", 2)
        self.assert_num_net(3)
        self.assert_num_wires(5)
        pyrtl.working_block().sanity_check()

    def test_concat(self):
        # concat's args are order dependent, therefore we need to check that we aren't
        # mangling them
        ins = [pyrtl.Input(5) for i in range(2)]
        outs = [pyrtl.Output(10) for i in range(2)]
        outs[0] <<= pyrtl.concat(ins[1], ins[0])
        outs[1] <<= pyrtl.concat(ins[0], ins[1])

        pyrtl.common_subexp_elimination()
        self.num_net_of_type("c", 2)
        self.num_net_of_type("w", 2)
        self.assert_num_net(4)
        self.assert_num_wires(6)
        pyrtl.working_block().sanity_check()

    def test_order_dependent_ops(self):
        # subtract, lt, gt simarlarly are order dependent. therefore we need to check
        # that we aren't mangling them
        for op, opcode in ((operator.sub, "-"), (operator.gt, ">"), (operator.lt, "<")):
            pyrtl.reset_working_block()
            ins = [pyrtl.Input(5) for i in range(2)]
            outs = [pyrtl.Output(10) for i in range(2)]
            outs[0] <<= op(ins[1], ins[0])
            outs[1] <<= op(ins[0], ins[1])

            pyrtl.common_subexp_elimination()
            self.num_net_of_type(opcode, 2)
            self.num_net_of_type("w", 2)
            pyrtl.working_block().sanity_check()

    def test_const_values_1(self):
        in_w = pyrtl.Input(5)
        out = pyrtl.Output(5)
        const = pyrtl.Const(23, 5)
        wire_1 = in_w + const
        wire_2 = in_w + const
        out <<= wire_1 | wire_2

        pyrtl.common_subexp_elimination()
        self.num_net_of_type("+", 1)
        self.num_net_of_type("w", 1)
        self.assert_num_net(4)  # because we cut off a bit after the add
        self.assert_num_wires(6)
        pyrtl.working_block().sanity_check()

    def test_const_values_2(self):
        in_w = pyrtl.Input(5)
        const = pyrtl.Const(23, 5)
        const_2 = pyrtl.Const(23, 5)
        _ = in_w + const
        _ = in_w + const_2

        pyrtl.common_subexp_elimination()
        self.num_net_of_type("+", 1)
        pyrtl.working_block().sanity_check()

    def test_const_different_bitwidth_1(self):
        in_w = pyrtl.Input(5)
        const = pyrtl.Const(23, 5)
        const_2 = pyrtl.Const(23, 6)
        _ = in_w + const
        _ = in_w + const_2

        pyrtl.common_subexp_elimination()
        self.num_net_of_type("+", 2)
        pyrtl.working_block().sanity_check()

    def test_no_elimination_of_different_const_bitwidths(self):
        # trying to merge const wires with different bitwidths together will cause
        # mismatches in bitwidths of certain wires
        const_1 = pyrtl.Const(3, 3)
        const_2 = pyrtl.Const(3, 5)
        out_1 = pyrtl.Output(5)
        out_2 = pyrtl.Output(5)
        out_1 <<= const_1 | const_1
        out_2 <<= const_2 | const_2
        pyrtl.common_subexp_elimination()

        self.num_net_of_type("|", 2)
        self.num_net_of_type("w", 2)
        self.assert_num_net(6)
        self.assert_num_wires(9)
        pyrtl.working_block().sanity_check()

    def test_multiple_elimination(self):
        ins = [pyrtl.Input(5) for i in range(3)]
        out_1 = pyrtl.Output(5)
        a = ins[0] ^ ins[1]
        b = ins[0] ^ ins[1]
        c = ins[0] ^ ins[1]

        out_1 <<= a | b | c
        pyrtl.common_subexp_elimination()
        self.num_net_of_type("^", 1)
        self.num_net_of_type("|", 2)
        pyrtl.working_block().sanity_check()

    def test_nested_elimination(self):
        ins = [pyrtl.Input(5) for i in range(3)]
        out_1 = pyrtl.Output(5)
        a = ins[0] ^ ins[0]
        b = ins[0] ^ ins[0]

        a2 = a & ins[2]
        b2 = b & ins[2]

        out_1 <<= a2 | b2
        pyrtl.common_subexp_elimination()
        self.num_net_of_type("^", 1)
        self.num_net_of_type("&", 1)
        pyrtl.working_block().sanity_check()


class TestSynthPasses(unittest.TestCase):
    in0: pyrtl.Input
    in1: pyrtl.Input
    out: pyrtl.Output

    def setUp(self):
        pyrtl.reset_working_block()
        self.in0 = pyrtl.Input(bitwidth=5, name="in0")
        self.in1 = pyrtl.Input(bitwidth=5, name="in1")
        self.out = pyrtl.Output(bitwidth=5, name="out")

    def check_synth(self, operation: Callable[[int, int], int]):
        """
        Simulates the current circuit with some test input pairs (in0, in1) and checks
        the outputs against the provided operation. Any synthesis/passes should be run
        before this gets called.

        :param operation: The operation to test against. This should be a lambda taking
            an input pair (in0, in1) and returning the expected output from this
            circuit.
        """
        sim = pyrtl.Simulation()

        values = [(1, 2), (4, 5), (7, 11)]
        """A list of input pairs (in0, in1) to test on."""

        for in0, in1 in values:
            expected_output = operation(in0, in1)
            sim.step({"in0": in0, "in1": in1})
            # compare simulation output to expected output
            self.assertEqual(
                sim.inspect("out"),
                expected_output,
                msg=f"Failed on inputs {in0} and {in1}",
            )

    def test_nand_synth_and(self):
        self.out <<= self.in0 & self.in1
        pyrtl.synthesize()
        pyrtl.nand_synth()
        self.check_synth(lambda a, b: a & b)

    def test_nand_synth_or(self):
        self.out <<= self.in0 | self.in1
        pyrtl.synthesize()
        pyrtl.nand_synth()
        self.check_synth(lambda a, b: a | b)

    def test_nand_synth_xor(self):
        self.out <<= self.in0 ^ self.in1
        pyrtl.synthesize()
        pyrtl.nand_synth()
        self.check_synth(lambda a, b: a ^ b)

    def test_nand_synth_adder(self):
        self.out <<= self.in0 + self.in1
        pyrtl.synthesize()
        pyrtl.nand_synth()
        self.check_synth(lambda a, b: a + b)

    def test_and_inverter_synth_and(self):
        self.out <<= self.in0 & self.in1
        pyrtl.synthesize()
        pyrtl.and_inverter_synth()
        self.check_synth(lambda a, b: a & b)

    def test_and_inverter_synth_or(self):
        self.out <<= self.in0 | self.in1
        pyrtl.synthesize()
        pyrtl.and_inverter_synth()
        self.check_synth(lambda a, b: a | b)

    def test_and_inverter_synth_xor(self):
        self.out <<= self.in0 ^ self.in1
        pyrtl.synthesize()
        pyrtl.and_inverter_synth()
        self.check_synth(lambda a, b: a ^ b)

    def test_and_inverter_synth_adder(self):
        self.out <<= self.in0 + self.in1
        pyrtl.synthesize()
        pyrtl.and_inverter_synth()
        self.check_synth(lambda a, b: a + b)


class TestSynthOptTiming(NetWireNumTestCases):
    def setUp(self):
        pyrtl.reset_working_block()

    def test_sanity_check(self):
        pass

    def everything_t_procedure(self, timing_val=None, opt_timing_val=None):
        # if there is a nondefault timing val supplied, then it will check to make sure
        # that the timing matches. this is a subprocess to do the synth and timing
        block = pyrtl.working_block()
        timing = pyrtl.TimingAnalysis(block)
        timing_max_length = timing.max_length()
        if timing_val is not None:
            self.assertEqual(timing_max_length, timing_val)
        _ = timing.critical_path(print_cp=False)

        pyrtl.synthesize()
        pyrtl.optimize()

        block = pyrtl.working_block()
        timing = pyrtl.TimingAnalysis(block)
        timing_max_length = timing.max_length()
        if opt_timing_val is not None:
            self.assertEqual(timing_max_length, opt_timing_val)
        _ = timing.critical_path(print_cp=False)

        pyrtl.and_inverter_synth()
        pyrtl.optimize()

        block = pyrtl.working_block()
        timing = pyrtl.TimingAnalysis(block)
        timing_max_length = timing.max_length()
        _ = timing.critical_path(print_cp=False)
        block = pyrtl.working_block()
        self.num_net_of_type("|", 0, block)
        self.num_net_of_type("^", 0, block)

        pyrtl.nand_synth()
        pyrtl.optimize()

        block = pyrtl.working_block()
        timing = pyrtl.TimingAnalysis(block)
        timing_max_length = timing.max_length()
        _ = timing.critical_path(print_cp=False)
        block.sanity_check()
        self.num_net_of_type("|", 0, block)
        self.num_net_of_type("^", 0, block)
        self.num_net_of_type("&", 0, block)

    def test_const_folding_complex_1(self):
        output = pyrtl.Output(bitwidth=3, name="output")
        counter = pyrtl.Register(bitwidth=3, name="counter")
        counter.next <<= counter + 1
        output <<= counter

        # just to check that something like this will run properly
        self.everything_t_procedure()

    def test_timing_basic_2(self):
        inwire, inwire2 = pyrtl.Input(bitwidth=1), pyrtl.Input(bitwidth=1)
        inwire3 = pyrtl.Input(bitwidth=1)
        tempwire, tempwire2 = pyrtl.WireVector(), pyrtl.WireVector()
        outwire = pyrtl.Output()

        tempwire <<= inwire | inwire2
        tempwire2 <<= ~tempwire
        outwire <<= tempwire2 & inwire3
        self.everything_t_procedure(252.3, 252.3)

    def test_timing_error(self):
        inwire, inwire2 = pyrtl.Input(bitwidth=1), pyrtl.Input(bitwidth=1)
        tempwire, tempwire2 = pyrtl.WireVector(1), pyrtl.WireVector(1)
        outwire = pyrtl.Output()

        tempwire <<= ~(inwire & tempwire2)
        tempwire2 <<= ~(inwire2 & tempwire)
        outwire <<= tempwire

        output = io.StringIO()
        sys.stdout = output
        with self.assertRaises(pyrtl.PyrtlError):
            pyrtl.synthesize()
            pyrtl.optimize()
            block = pyrtl.working_block()
            _timing = pyrtl.TimingAnalysis(block)
        sys.stdout = sys.__stdout__
        self.assertTrue(output.getvalue().startswith("Loop found:"))

    def test_wirevector_1(self):
        inwire = pyrtl.Input(bitwidth=1)
        tempwire0, tempwire1 = (
            pyrtl.WireVector(bitwidth=1),
            pyrtl.WireVector(bitwidth=1),
        )
        tempwire2 = pyrtl.WireVector(bitwidth=1)
        outwire = pyrtl.Output()

        tempwire0 <<= inwire
        tempwire1 <<= tempwire0
        tempwire2 <<= tempwire1
        outwire <<= ~tempwire2
        self.everything_t_procedure(48.5, 48.5)
        block = pyrtl.working_block()
        self.assert_num_net(2, block)

    def test_combo_1(self):
        inwire, inwire2 = pyrtl.Input(bitwidth=1), pyrtl.Input(bitwidth=1)
        tempwire, tempwire2 = pyrtl.WireVector(), pyrtl.WireVector()
        inwire3 = pyrtl.Input(bitwidth=1)
        outwire = pyrtl.Output()

        tempwire <<= inwire | inwire2
        tempwire2 <<= ~tempwire
        outwire <<= tempwire2 & inwire3
        self.everything_t_procedure(252.3, 252.3)

    def test_adder(self):
        inwire1, inwire2 = pyrtl.Input(bitwidth=3), pyrtl.Input(bitwidth=3)
        outwire = pyrtl.Output(bitwidth=4)

        outwire <<= inwire1 + inwire2
        self.everything_t_procedure()

    def test_all_mem_1(self):
        readAdd1, readAdd2, writeAdd1, writeAdd2, readData1, readData2 = (
            pyrtl.Input(bitwidth=3) for i in range(6)
        )

        dataOut = pyrtl.Output(bitwidth=3)

        memory = pyrtl.MemBlock(3, 3, asynchronous=True)

        memory[readAdd1 & readAdd2] <<= readData1 ^ readData2
        dataOut <<= memory[writeAdd1 | writeAdd2]
        self.everything_t_procedure()


class TestConcatAndSelectSimplification(unittest.TestCase):
    def setUp(self):
        pyrtl.reset_working_block()

    def test_two_way_concat(self):
        i = pyrtl.Const(0b1100)
        j = pyrtl.Const(0b011, bitwidth=3)
        k = pyrtl.Const(0b100110)
        o = pyrtl.Output(13, "o")
        o <<= pyrtl.concat(i, j, k)

        block = pyrtl.working_block()
        concat_nets = list(block.logic_subset(op="c"))
        self.assertEqual(len(concat_nets), 1)
        self.assertEqual(concat_nets[0].args, (i, j, k))

        pyrtl.two_way_concat()

        concat_nets = list(block.logic_subset(op="c"))
        self.assertEqual(len(concat_nets), 2)
        upper_concat = next(n for n in concat_nets if i is n.args[0])
        lower_concat = next(n for n in concat_nets if k is n.args[1])
        self.assertNotEqual(upper_concat, lower_concat)
        self.assertEqual(upper_concat.args, (i, j))
        self.assertEqual(lower_concat.args, (upper_concat.dests[0], k))

        sim = pyrtl.Simulation()
        sim.step()
        self.assertEqual(sim.inspect("o"), 0b1100011100110)

    def test_one_bit_selects(self):
        a = pyrtl.Const(0b101101001101)
        b = pyrtl.Output(6, "b")
        b <<= a[::2]  # bits 0, 2, 4, 6, 8, and 10 of wire a

        block = pyrtl.working_block()
        select_nets = list(block.logic_subset(op="s"))
        self.assertEqual(len(select_nets), 1)
        self.assertEqual(tuple(select_nets[0].op_param), (0, 2, 4, 6, 8, 10))

        pyrtl.one_bit_selects()

        select_nets = list(block.logic_subset(op="s"))
        for net in select_nets:
            indices = net.op_param
            self.assertEqual(len(indices), 1)

        sim = pyrtl.Simulation()
        sim.step()
        self.assertEqual(sim.inspect("b"), 0b00011011)


class TestDirectlyConnectedOutputs(unittest.TestCase):
    def setUp(self):
        pyrtl.reset_working_block()

    def test_single_output(self):
        i, j = pyrtl.input_list("i/4 j/4")
        o = pyrtl.Output(8, "o")
        o <<= i * j

        self.assertEqual(len(pyrtl.working_block().logic), 2)
        # Includes intermediary 'w' wire
        self.assertEqual(len(pyrtl.working_block().wirevector_set), 4)
        pyrtl.direct_connect_outputs()
        self.assertEqual(len(pyrtl.working_block().logic), 1)
        self.assertEqual(pyrtl.working_block().wirevector_set, {i, j, o})

    def test_single_output_simulates_correctly(self):
        i, ivals = utils.an_input_and_vals(4, name="i")
        j, jvals = utils.an_input_and_vals(4, name="j")
        o = pyrtl.Output(8, "o")
        o <<= i * j

        pyrtl.direct_connect_outputs()
        true_result = [x * y for x, y in zip(ivals, jvals)]
        sim_result = utils.sim_and_ret_out(o, [i, j], [ivals, jvals])
        self.assertEqual(true_result, sim_result)

    def test_several_outputs(self):
        i, j = pyrtl.input_list("i/2 j/2")
        o, p, q = pyrtl.output_list("o p q")
        o <<= i * j
        w = i + 2
        p <<= w
        q <<= ~w

        self.assertEqual(len(pyrtl.working_block().logic), 9)
        self.assertEqual(len(pyrtl.working_block().logic_subset(op="w")), 3)
        pyrtl.direct_connect_outputs()
        self.assertEqual(len(pyrtl.working_block().logic), 6)
        self.assertEqual(len(pyrtl.working_block().logic_subset(op="w")), 0)

    def test_several_outputs_simulates_correctly(self):
        i, j = pyrtl.input_list("i/2 j/2")
        o, p, q = pyrtl.output_list("o p q")
        o <<= i * j
        w = i + 2
        p <<= w
        q <<= ~w

        inputs = [(0, 1), (1, 0), (2, 3), (3, 0), (1, 3)]
        trace_pre = pyrtl.SimulationTrace()
        sim = pyrtl.Simulation(tracer=trace_pre)
        for x, y in inputs:
            inp_map = {"i": x, "j": y}
            sim.step(inp_map)

        pyrtl.direct_connect_outputs()

        trace_post = pyrtl.SimulationTrace()
        sim = pyrtl.Simulation(tracer=trace_post)
        for x, y in inputs:
            inp_map = {"i": x, "j": y}
            sim.step(inp_map)

        self.assertEqual(trace_pre.trace, trace_post.trace)

    def test_some_outputs_unaffected(self):
        i = pyrtl.Input(2, "i")
        o, p, q = pyrtl.output_list("o/4 p/4 q/2")
        w = i * 2
        o <<= w
        p <<= w
        q <<= ~i

        src_nets, _ = pyrtl.working_block().net_connections()
        self.assertEqual(src_nets[o].op, "w")
        self.assertEqual(src_nets[p].op, "w")
        self.assertEqual(src_nets[q].op, "w")
        self.assertEqual(len(pyrtl.working_block().logic), 5)

        pyrtl.direct_connect_outputs()
        src_nets, _ = pyrtl.working_block().net_connections()
        self.assertEqual(src_nets[o].op, "w")
        self.assertEqual(src_nets[p].op, "w")
        self.assertEqual(src_nets[q].op, "~")
        self.assertEqual(len(pyrtl.working_block().logic), 4)


class TestTwoWayFanout(unittest.TestCase):
    def setUp(self):
        pyrtl.reset_working_block()

    def check_all_leq_two(self):
        for w in pyrtl.working_block().wirevector_subset(exclude=(pyrtl.Output)):
            self.assertLessEqual(pyrtl.fanout(w), 2)

    def test_two_way_fanout_small_design(self):
        i = pyrtl.Input(1, "i")
        o, p, q = pyrtl.output_list("o p q")
        o <<= ~i
        p <<= i
        q <<= i & 0

        self.assertEqual(pyrtl.fanout(i), 3)
        pyrtl.two_way_fanout()
        self.check_all_leq_two()

    def test_two_way_fanout_medium_design(self):
        i = pyrtl.Input(1, "i")
        o, p = pyrtl.output_list("o p")
        w = i & 0
        x = (w & w) ^ w
        o <<= x
        p <<= w

        self.assertEqual(pyrtl.fanout(w), 4)
        pyrtl.two_way_fanout()
        self.check_all_leq_two()

    def test_two_way_fanout_large_design(self):
        i, j = pyrtl.input_list("i/1 j/2")
        o, p, q, r = pyrtl.output_list("o p q r")
        o <<= ~i
        p <<= i * j
        q <<= i & 0
        r <<= i - 3

        self.assertEqual(pyrtl.fanout(i), 4)
        pyrtl.two_way_fanout()
        self.check_all_leq_two()


if __name__ == "__main__":
    unittest.main()



import pyrtl


def match_bitwidth(*args: pyrtl.WireVector):
    # TODO: allow for custom bit extension functions
    """Matches the bitwidth of all of the input arguments.

    .. WARNING::

        Use :func:`.match_bitwidth` instead.

    :param args: input arguments

    :return: tuple of ``args`` in order with extended bits
    """
    return pyrtl.match_bitwidth(*args)


def partition_wire(
    wire: pyrtl.WireVector, partition_size: int
) -> list[pyrtl.WireVector]:
    """Partitions a wire into a list of ``N`` wires of size ``partition_size``.

    The ``wire``'s bitwidth must be evenly divisible by ``partition_size``.

    .. WARNING::

        Use :func:`.wire_matrix` or :func:`.chop` instead.

    :param wire: Wire to partition.
    :param partition_size: Integer representing size of each partition.
    """
    if len(wire) % partition_size != 0:
        msg = (
            f"Wire {wire} cannot be evenly partitioned into items of size "
            f"{partition_size}"
        )
        raise pyrtl.PyrtlError(msg)
    return [
        wire[offset : offset + partition_size]
        for offset in range(0, len(wire), partition_size)
    ]


def str_to_int_array(string, base=16):
    """
    Converts a string to an array of integer values according to the base specified (int
    numbers must be whitespace delimited).

    Example: ``"13 a3 3c" => [0x13, 0xa3, 0x3c]``

    .. WARNING::

        Use a :class:`list` comprehension instead::

            >>> hex_string = "13 a3 3c"

            >>> int_list = [int(s, base=16) for s in hex_string.split()]

            >>> int_list
            [19, 163, 60]
            >>> [hex(i) for i in int_list]
            ['0x13', '0xa3', '0x3c']


    :return: [int]
    """
    int_strings = string.split()
    return [int(int_str, base) for int_str in int_strings]


def twos_comp_repr(val: int, bitwidth: int) -> int:
    """Converts a value to its two's-complement (positive) integer representation using
    a given bitwidth (only converts the value if it is negative).

    .. WARNING::

        Use :func:`.infer_val_and_bitwidth` instead.

    :param val: Integer literal to convert to two's complement
    :param bitwidth: Size of val in bits
    """
    correctbw = abs(val).bit_length() + 1
    if bitwidth < correctbw:
        msg = "please choose a larger target bitwidth"
        raise pyrtl.PyrtlError(msg)
    if val >= 0:
        return val
    return (~abs(val) & (2**bitwidth - 1)) + 1  # flip the bits and add one


def rev_twos_comp_repr(val: int, bitwidth: int) -> int:
    """Takes a two's-complement represented value and converts it to a signed integer
    based on the provided ``bitwidth``.

    .. WARNING::

        Use :func:`.val_to_signed_integer` instead.
    """
    valbl = val.bit_length()
    if bitwidth < val.bit_length() or val == 2 ** (bitwidth - 1):
        msg = "please choose a larger target bitwidth"
        raise pyrtl.PyrtlError(msg)
    if bitwidth == valbl:  # MSB is a 1, value is negative
        return -(
            (~val & (2**bitwidth - 1)) + 1
        )  # flip the bits, add one, and make negative
    return val


def _shifted_reg_next(reg: pyrtl.Register, direct: str, num: int = 1):
    """Creates a shifted :attr:`.Register.next` property for shifted (left or right)
    register.

    Use: ``myReg.next = shifted_reg_next(myReg, 'l', 4)``

    .. WARNING::

        Use :func:`.shift_left_logical` or :func:`.shift_right_logical` instead.

    :param direct: Direction of shift, either ``l`` or ``r``.
    :param num: Number of bit positions to shift.

    :return: Register containing reg's (shifted) next state
    """
    if direct == "l":
        if num >= len(reg):
            return 0
        return pyrtl.concat(reg, pyrtl.Const(0, num))
    if direct == "r":
        if num >= len(reg):
            return 0
        return reg[num:]
    msg = "direction must be specified with 'direct' parameter as either 'l' or 'r'"
    raise pyrtl.PyrtlError(msg)



"""
Basic integer addition is defined in PyRTL's core library, see:

- :meth:`.WireVector.__add__` for unsigned integer addition.

- :func:`.signed_add` for signed integer addition.

The functions below provide more complex alternatives.
"""

import itertools
import math
from typing import Callable

import pyrtl


def kogge_stone(
    a: pyrtl.WireVector, b: pyrtl.WireVector, cin: pyrtl.wire.WireVectorLike = 0
) -> pyrtl.WireVector:
    """Creates a Kogge-Stone adder given two inputs.

    The Kogge-Stone adder is a fast tree-based adder with `O(log(n))` propagation delay,
    useful for performance critical designs. However, it has `O(n log(n))` area usage,
    and large fan out.

    .. doctest only::

        >>> import pyrtl
        >>> pyrtl.reset_working_block()

    Example::

        >>> a = pyrtl.Input(name="a", bitwidth=4)
        >>> b = pyrtl.Input(name="b", bitwidth=4)
        >>> output = pyrtl.Output(name="output")

        >>> output <<= pyrtl.rtllib.adders.kogge_stone(a, b)

        >>> sim = pyrtl.Simulation()
        >>> sim.step(provided_inputs={"a": 2, "b": 3})
        >>> sim.inspect("output")
        5

    :param a: A :class:`.WireVector` to add up. Bitwidths don't need to match.
    :param b: A :class:`.WireVector` to add up. Bitwidths don't need to match.
    :param cin: An optional 1-bit carry-in :class:`.WireVector`. Can be any type that
        can be coerced to :class:`.WireVector` by :func:`.as_wires`.

    :return: A :class:`.WireVector` representing the output of the adder.
    """
    a, b = pyrtl.match_bitwidth(a, b)

    prop_orig = a ^ b
    prop_bits = list(prop_orig)
    gen_bits = list(a & b)
    prop_dist = 1

    # creation of the carry calculation
    while prop_dist < len(a):
        for i in reversed(range(prop_dist, len(a))):
            prop_old = prop_bits[i]
            gen_bits[i] = gen_bits[i] | (prop_old & gen_bits[i - prop_dist])
            if i >= prop_dist * 2:  # to prevent creating unnecessary nets and wires
                prop_bits[i] = prop_old & prop_bits[i - prop_dist]
        prop_dist *= 2

    # assembling the result of the addition. preparing the cin (and conveniently
    # shifting the gen bits)
    gen_bits.insert(0, pyrtl.as_wires(cin))
    return pyrtl.concat_list(gen_bits) ^ prop_orig


def one_bit_add(a, b, cin=0):
    return pyrtl.concat(*_one_bit_add_no_concat(a, b, cin))


def _one_bit_add_no_concat(a, b, cin=0):
    cin = pyrtl.as_wires(cin)  # to make sure that an int cin doesn't break things
    assert len(a) == len(b) == len(cin) == 1
    sum = a ^ b ^ cin
    cout = a & b | a & cin | b & cin
    return cout, sum


def half_adder(a, b):
    assert len(a) == len(b) == 1
    sum = a ^ b
    cout = a & b
    return cout, sum


def ripple_add(a, b, cin=0):
    if len(a) < len(b):  # make sure that b is the shorter wire
        b, a = a, b
    cin = pyrtl.as_wires(cin)
    if len(a) == 1:
        return one_bit_add(a, b, cin)
    ripplecarry = one_bit_add(a[0], b[0], cin)
    if len(b) == 1:
        msbits = ripple_half_add(a[1:], ripplecarry[1])
    else:
        msbits = ripple_add(a[1:], b[1:], ripplecarry[1])
    return pyrtl.concat(msbits, ripplecarry[0])


def ripple_half_add(a, cin=0):
    cin = pyrtl.as_wires(cin)
    if len(a) == 1:
        return pyrtl.concat(*half_adder(a, cin))
    ripplecarry = half_adder(a[0], cin)
    msbits = ripple_half_add(a[1:], ripplecarry[0])
    return pyrtl.concat(msbits, ripplecarry[1])


def carrysave_adder(
    a: pyrtl.WireVector,
    b: pyrtl.WireVector,
    c: pyrtl.WireVector,
    final_adder: Callable = ripple_add,
) -> pyrtl.WireVector:
    """Adds three :class:`WireVectors<.WireVector>` up in an efficient manner.

    .. doctest only::

        >>> import pyrtl
        >>> pyrtl.reset_working_block()

    Example::

        >>> a = pyrtl.Input(name="a", bitwidth=4)
        >>> b = pyrtl.Input(name="b", bitwidth=4)
        >>> c = pyrtl.Input(name="c", bitwidth=4)
        >>> output = pyrtl.Output(name="output")

        >>> output <<= pyrtl.rtllib.adders.carrysave_adder(a, b, c)

        >>> sim = pyrtl.Simulation()
        >>> sim.step(provided_inputs={"a": 2, "b": 3, "c": 4})
        >>> sim.inspect("output")
        9

    :param a: A :class:`.WireVector` to add up. Bitwidths don't need to match.
    :param b: A :class:`.WireVector` to add up. Bitwidths don't need to match.
    :param c: A :class:`.WireVector` to add up. Bitwidths don't need to match.
    :param final_adder: The adder to use for the final addition.

    :return: A :class:`.WireVector` with bitwidth equal to the longest input, plus 2.
    """
    a, b, c = pyrtl.match_bitwidth(a, b, c)
    partial_sum = a ^ b ^ c
    shift_carry = (a | b) & (a | c) & (b | c)
    return pyrtl.concat(final_adder(partial_sum[1:], shift_carry), partial_sum[0])


def cla_adder(
    a: pyrtl.WireVector,
    b: pyrtl.WireVector,
    cin: pyrtl.WireVector = 0,
    la_unit_len: int = 4,
) -> pyrtl.WireVector:
    """Carry Look-Ahead Adder.

    A Carry Look-Ahead Adder is an adder that is faster than :func:`ripple_add`, as it
    calculates the carry bits faster. It is not as fast as :func:`kogge_stone`, but uses
    less area.

    .. doctest only::

        >>> import pyrtl
        >>> pyrtl.reset_working_block()

    Example::

        >>> a = pyrtl.Input(name="a", bitwidth=4)
        >>> b = pyrtl.Input(name="b", bitwidth=4)
        >>> output = pyrtl.Output(name="output")

        >>> output <<= pyrtl.rtllib.adders.cla_adder(a, b)

        >>> sim = pyrtl.Simulation()
        >>> sim.step(provided_inputs={"a": 2, "b": 3})
        >>> sim.inspect("output")
        5

    :param a: A :class:`.WireVector` to add up. Bitwidths don't need to match.
    :param b: A :class:`.WireVector` to add up. Bitwidths don't need to match.
    :param cin: A 1-bit carry-in :class:`.WireVector`.
    :param la_unit_len: The length of input that every unit processes.

    :return: A :class:`.WireVector` representing the output of the adder.
    """
    a, b = pyrtl.match_bitwidth(a, b)
    if len(a) <= la_unit_len:
        sum, cout = _cla_adder_unit(a, b, cin)
        return pyrtl.concat(cout, sum)
    sum, cout = _cla_adder_unit(a[0:la_unit_len], b[0:la_unit_len], cin)
    msbits = cla_adder(a[la_unit_len:], b[la_unit_len:], cout, la_unit_len)
    return pyrtl.concat(msbits, sum)


def _cla_adder_unit(a, b, cin):
    """
    Carry generation and propogation signals will be calculated only using the inputs;
    their values don't rely on the sum. Every unit generates a cout signal which is used
    as cin for the next unit.
    """
    gen = a & b
    prop = a ^ b
    assert len(prop) == len(gen)

    carry = [gen[0] | prop[0] & cin]
    sum_bit = prop[0] ^ cin

    cur_gen = gen[0]
    cur_prop = prop[0]
    for i in range(1, len(prop)):
        cur_gen = gen[i] | (prop[i] & cur_gen)
        cur_prop = cur_prop & prop[i]
        sum_bit = pyrtl.concat(prop[i] ^ carry[i - 1], sum_bit)
        carry.append(gen[i] | (prop[i] & carry[i - 1]))
    cout = cur_gen | (cur_prop & cin)
    return sum_bit, cout


def wallace_reducer(
    wire_array_2: list[list[pyrtl.WireVector]],
    result_bitwidth: int,
    final_adder: Callable = kogge_stone,
) -> pyrtl.WireVector:
    """The reduction and final adding part of a dada tree.

    Useful for adding many numbers together with :func:`fast_group_adder`. The use of
    single bitwidth wires allows for additional flexibility.

    :param wire_array_2: An array of arrays of single bitwidth
        :class:`WireVectors<.WireVector>`.
    :param result_bitwidth: Bitwidth of the resulting wire. Used to eliminate
        unnecessary wires.
    :param final_adder: The adder used for the final addition.

    :return: :class:`.WireVector` with :attr:`~.WireVector.bitwidth`
             ``result_bitwidth``.
    """
    # verification that the wires are actually wirevectors of length 1
    for wire_set in wire_array_2:
        for a_wire in wire_set:
            if not isinstance(a_wire, pyrtl.WireVector) or len(a_wire) != 1:
                msg = (
                    f"The item {a_wire} is not a valid element for the wire_array_2. "
                    "It must be a WireVector of bitwidth 1"
                )
                raise pyrtl.PyrtlError(msg)

    while not all(len(i) <= 2 for i in wire_array_2):
        deferred = [[] for weight in range(result_bitwidth + 1)]
        for i, w_array in enumerate(
            wire_array_2
        ):  # Start with low weights and start reducing
            while len(w_array) >= 3:
                cout, sum = _one_bit_add_no_concat(*(w_array.pop(0) for j in range(3)))
                deferred[i].append(sum)
                deferred[i + 1].append(cout)

            if len(w_array) == 2:
                cout, sum = half_adder(*w_array)
                deferred[i].append(sum)
                deferred[i + 1].append(cout)
            else:
                deferred[i].extend(w_array)
        wire_array_2 = deferred[:result_bitwidth]

    # At this stage in the multiplication we have only 2 wire vectors left.
    # now we need to add them up
    result = _sparse_adder(wire_array_2, final_adder)
    if len(result) > result_bitwidth:
        return result[:result_bitwidth]
    return result


def dada_reducer(
    wire_array_2: list[list[pyrtl.WireVector]],
    result_bitwidth: int,
    final_adder: Callable = kogge_stone,
) -> pyrtl.WireVector:
    """The reduction and final adding part of a dada tree.

    Useful for adding many numbers together with :func:`fast_group_adder`. The use of
    single bitwidth wires allows for additional flexibility.

    :param wire_array_2: An array of arrays of single bitwidth
        :class:`WireVectors<.WireVector>`.
    :param result_bitwidth: Bitwidth of the resulting wire. Used to eliminate
        unnecessary wires.
    :param final_adder: The adder used for the final addition.

    :return: :class:`.WireVector` with :attr:`~.WireVector.bitwidth`
             ``result_bitwidth``.
    """
    # verification that the wires are actually wirevectors of length 1
    for wire_set in wire_array_2:
        for a_wire in wire_set:
            if not isinstance(a_wire, pyrtl.WireVector) or len(a_wire) != 1:
                msg = (
                    f"The item {a_wire} is not a valid element for the wire_array_2. "
                    "It must be a WireVector of bitwidth 1"
                )
                raise pyrtl.PyrtlError(msg)

    max_width = max(len(i) for i in wire_array_2)
    reduction_schedule = [2]
    while reduction_schedule[-1] <= max_width:
        reduction_schedule.append(int(reduction_schedule[-1] * 3 / 2))

    for reduction_target in reversed(reduction_schedule[:-1]):
        deferred = [[] for weight in range(result_bitwidth + 1)]
        for i, w_array in enumerate(
            wire_array_2
        ):  # Start with low weights and start reducing
            while len(w_array) + len(deferred[i]) > reduction_target:
                if len(w_array) + len(deferred[i]) - reduction_target >= 2:
                    cout, sum = _one_bit_add_no_concat(
                        *(w_array.pop(0) for j in range(3))
                    )
                    deferred[i].append(sum)
                    deferred[i + 1].append(cout)
                else:
                    cout, sum = half_adder(*(w_array.pop(0) for j in range(2)))
                    deferred[i].append(sum)
                    deferred[i + 1].append(cout)
            deferred[i].extend(w_array)
            if len(deferred[i]) > reduction_target:
                msg = "Expected that the code would be able to reduce more wires"
                raise pyrtl.PyrtlError(msg)
        wire_array_2 = deferred[:result_bitwidth]

    # At this stage in the multiplication we have only 2 wire vectors left.
    # now we need to add them up
    result = _sparse_adder(wire_array_2, final_adder)
    if len(result) > result_bitwidth:
        return result[:result_bitwidth]
    return result


def _sparse_adder(wire_array_2, adder):
    result = []
    for single_w_index in range(len(wire_array_2)):
        if (
            len(wire_array_2[single_w_index]) == 2
        ):  # Check if the two wire vectors overlap yet
            break
        result.append(wire_array_2[single_w_index][0])

    wires_to_zip = wire_array_2[single_w_index:]
    add_wires = tuple(itertools.zip_longest(*wires_to_zip, fillvalue=pyrtl.Const(0)))
    adder_result = adder(
        pyrtl.concat_list(add_wires[0]), pyrtl.concat_list(add_wires[1])
    )
    return pyrtl.concat(adder_result, *reversed(result))


"""
Some adders that utilize these tree reducers
"""


def fast_group_adder(
    wires_to_add: list[pyrtl.WireVector],
    reducer: Callable = wallace_reducer,
    final_adder: Callable = kogge_stone,
):
    """A generalization of :func:`carrysave_adder`, ``fast_group_adder`` is designed to
    add many numbers together in a both area and time efficient manner. Uses a tree
    reducer to achieve this performance.

    The length of the result is::

        max(len(w) for w in wires_to_add) + ceil(len(wires_to_add))

    .. doctest only::

        >>> import pyrtl
        >>> pyrtl.reset_working_block()

    Example::

        >>> wires_to_add = [pyrtl.Const(n) for n in range(10)]
        >>> output = pyrtl.Output(name="output")

        >>> output <<= pyrtl.rtllib.adders.fast_group_adder(wires_to_add)

        >>> sim = pyrtl.Simulation()
        >>> sim.step()
        >>> sim.inspect("output")
        45
        >>> sum(range(10))
        45


    :param wires_to_add: A :class:`list` of :class:`WireVectors<.WireVector>` to add.
    :param reducer: The tree reducer to use. See :func:`wallace_reducer` and
        :func:`dada_reducer`.
    :param final_adder: The two value adder to use at the end.

    :return: A :class:`.WireVector` with the result of the addition.
    """
    longest_wire_len = max(len(w) for w in wires_to_add)
    result_bitwidth = longest_wire_len + math.ceil(math.log2(len(wires_to_add)))

    bits = [[] for i in range(longest_wire_len)]

    for wire in wires_to_add:
        for bit_loc, bit in enumerate(wire):
            bits[bit_loc].append(bit)

    return reducer(bits, result_bitwidth, final_adder)



"""
Basic shifting is defined in PyRTL's core library, see:

- :func:`.shift_left_logical`

- :func:`.shift_right_logical`

- :func:`.shift_right_arithmetic`

:func:`barrel_shifter` should only be used when more complex shifting behavior is
required.
"""

from enum import IntEnum

from pyrtl.wire import WireVector, WireVectorLike


class Direction(IntEnum):
    """Assigns names to each shift direction, to improve code readability."""

    RIGHT = 0
    LEFT = 1


def barrel_shifter(
    bits_to_shift: WireVector,
    bit_in: WireVectorLike,
    direction: WireVectorLike,
    shift_dist: WireVector,
    wrap_around=0,
) -> WireVector:
    """Create a barrel shifter.

    .. doctest only::

        >>> import pyrtl
        >>> pyrtl.reset_working_block()

    Example::

        >>> bits_to_shift = pyrtl.Input(name="input", bitwidth=8)
        >>> shift_dist = pyrtl.Input(name="shift_dist", bitwidth=3)
        >>> output = pyrtl.Output(name="output")

        >>> output <<= pyrtl.rtllib.barrel.barrel_shifter(
        ...     bits_to_shift,
        ...     bit_in=1,
        ...     direction=pyrtl.rtllib.barrel.Direction.RIGHT,
        ...     shift_dist=shift_dist)

        >>> sim = pyrtl.Simulation()
        >>> sim.step(provided_inputs={"input": 0x55, "shift_dist": 4})
        >>> hex(sim.inspect("output"))
        '0xf5'

    :param bits_to_shift: :class:`.WireVector` with the value to shift.
    :param bit_in: A 1-bit :class:`.WireVector` representing the value to shift in.
    :param direction: A one bit :class:`.WireVector` representing the shift direction
        (``0`` = shift right, ``1`` = shift left). If ``direction`` is constant, use
        :class:`Direction` to improve code readability (``direction=Direction.RIGHT``
        instead of ``direction=0``).
    :param shift_dist: :class:`.WireVector` representing the amount to shift.
    :param wrap_around: ****currently not implemented****

    :return: The shifted :class:`.WireVector`.
    """
    from pyrtl import concat, select  # just for readability

    if wrap_around != 0:
        raise NotImplementedError

    # Implement with logN stages pyrtl.muxing between shifted and un-shifted values
    final_width = len(bits_to_shift)
    val = bits_to_shift
    append_val = bit_in

    for i in range(len(shift_dist)):
        shift_amt = pow(2, i)  # stages shift 1,2,4,8,...
        if shift_amt < final_width:
            newval = select(
                direction,
                concat(val[:-shift_amt], append_val),  # shift left
                concat(append_val, val[shift_amt:]),
            )  # shift right
            val = select(
                shift_dist[i],
                truecase=newval,  # if bit of shift is 1, do the shift
                falsecase=val,
            )  # otherwise, don't
            # the value to append grows exponentially, but is capped at full width
            append_val = concat(append_val, append_val)[:final_width]
        else:
            # if we are shifting this much, all the data is gone
            val = select(
                shift_dist[i],
                truecase=append_val,  # if bit of shift is 1, do the shift
                falsecase=val,
            )  # otherwise, don't

    return val



import unittest

import pyrtl
import pyrtl.rtllib.testingutils as utils
from pyrtl.rtllib import libutils


class TestPartitionWire(unittest.TestCase):
    def test_successful_partition(self):
        w = pyrtl.WireVector(24)
        partitioned_w = libutils.partition_wire(w, 4)
        self.assertEqual(len(partitioned_w), 6)
        for wire in partitioned_w:
            self.assertIsInstance(wire, pyrtl.WireVector)

    def test_failing_partition(self):
        w = pyrtl.WireVector(14)
        with self.assertRaises(pyrtl.PyrtlError):
            _ = libutils.partition_wire(w, 4)

    def test_partition_sim(self):
        pyrtl.reset_working_block()
        wires, vals = utils.make_inputs_and_values(exact_bitwidth=32, num_wires=1)
        out_wires = [pyrtl.Output(8, "o" + str(i)) for i in range(4)]
        partitioned_w = libutils.partition_wire(wires[0], 8)
        for p_wire, o_wire in zip(partitioned_w, out_wires):
            o_wire <<= p_wire

        out_vals = utils.sim_and_ret_outws(wires, vals)
        partitioned_vals = [
            [(val >> i) & 0xFF for i in (0, 8, 16, 24)] for val in vals[0]
        ]
        true_vals = tuple(zip(*partitioned_vals))
        for index, wire in enumerate(out_wires):
            self.assertEqual(tuple(out_vals[wire.name]), true_vals[index])


class TestStringConversion(unittest.TestCase):
    def test_simple_conversion(self):
        self.assertEqual([0xA7, 0x23], libutils.str_to_int_array("a7 23"))

    def test_binary_conversion(self):
        result = libutils.str_to_int_array("0100 0110 010", base=2)
        self.assertEqual(result, [4, 6, 2])

    def test_empty(self):
        result = libutils.str_to_int_array("")
        self.assertEqual(result, [])

    def test_multiline(self):
        text = """
        374 1c
        a
        34 76"""
        result = libutils.str_to_int_array(text)
        self.assertEqual([0x374, 0x1C, 0xA, 0x34, 0x76], result)

    def test_invalid_str(self):
        with self.assertRaises(ValueError):
            libutils.str_to_int_array("hello")

    def test_invalid_bin_str(self):
        with self.assertRaises(ValueError):
            libutils.str_to_int_array("0313", 2)

    def test_no_override(self):
        with self.assertRaises(ValueError):
            libutils.str_to_int_array("0x0313", 2)


class TestTwosComp(unittest.TestCase):
    def setUp(self):
        pyrtl.reset_working_block()
        self.in1, self.in2 = (pyrtl.Input(8, "in" + str(i)) for i in range(1, 3))
        self.out = pyrtl.Output(9, "out")

    def test_inverse_functionality(self):
        for i in range(20):
            self.assertEqual(
                i * 3,
                libutils.rev_twos_comp_repr(libutils.twos_comp_repr(i * 3, 16), 16),
            )

    def test_low_bw_error(self):
        with self.assertRaises(pyrtl.PyrtlError):
            libutils.twos_comp_repr(-40, 6)
        with self.assertRaises(pyrtl.PyrtlError):
            libutils.rev_twos_comp_repr(88, 6)
        with self.assertRaises(pyrtl.PyrtlError):
            libutils.rev_twos_comp_repr(8, 4)

    def test_twos_comp_sim(self):
        self.out <<= self.in1 + self.in2
        sim = pyrtl.Simulation()
        for i in range(10):
            sim.step({"in1": i, "in2": libutils.twos_comp_repr(-2 * i, 8)})
            self.assertEqual(-i, libutils.rev_twos_comp_repr(sim.inspect("out"), 8))


if __name__ == "__main__":
    unittest.main()



# # Example 7: Timing Analysis and Optimization
#
# After building a circuit, one might want to analyze or simplify the hardware. PyRTL
# provides this functionality, as demonstrated by this example.

import pyrtl

# ## Part 1: Timing Analysis
#
# Timing and area usage are key considerations of any hardware block that one makes.
# PyRTL provides functions to do these operations.
#
# Creating a sample hardware block
pyrtl.reset_working_block()
const_wire = pyrtl.Const(6, bitwidth=4)
in_wire2 = pyrtl.Input(bitwidth=4, name="input2")
out_wire = pyrtl.Output(bitwidth=5, name="output")
out_wire <<= const_wire + in_wire2

# Now we will do the timing analysis as well as print out the critical path
#
# Generating timing analysis information
print("Pre Synthesis:")
timing = pyrtl.TimingAnalysis()
timing.print_max_length()

# We are also able to print out the critical paths as well as get them back as an array.
critical_path_info = timing.critical_path()

# ## Part 2: Area Analysis
#
# PyRTL also provides estimates for the area that would be used up if the circuit was
# printed as an ASIC.
logic_area, mem_area = pyrtl.area_estimation(tech_in_nm=65)
est_area = logic_area + mem_area
print("Estimated Area of block", est_area, "sq mm")


# ## Part 3: Synthesis
#
# Synthesis is the operation of reducing the circuit into simpler components. The base
# synthesis function breaks down the more complex logic operations into logic gates
# (keeping registers and memories intact) as well as reduces all combinatorial logic
# into operations that only use 1-bitwidth wires.
#
# This synthesis allows for PyRTL to make optimizations to the net structure as well as
# prepares it for further transformations on the PyRTL toolchain.
pyrtl.synthesize()

print("\nPre Optimization:")
timing = pyrtl.TimingAnalysis()
timing.print_max_length()
for net in pyrtl.working_block().logic:
    print(str(net))


# ## Part 4: Optimization
#
# PyRTL has functions built-in to eliminate unnecessary logic from the circuit. These
# functions are all done with a simple call:
_ = pyrtl.optimize()

# Now to see the difference
print("\nPost Optimization:")
timing = pyrtl.TimingAnalysis()
timing.print_max_length()

for net in pyrtl.working_block().logic:
    print(str(net))

# As we can see, the number of nets in the circuit was drastically reduced by the
# optimization algorithm.



# # Example 1.1: Working with signed integers.
#
# This example demonstrates:
# * Correct addition of signed integers with `signed_add()`.
# * Displaying signed integers in traces with `val_to_signed_integer()`.
#
# Signed integers are represented in two's complement.
# https://en.wikipedia.org/wiki/Two%27s_complement
import pyrtl

# ## Let's start with unsigned addition.
#
# Add all combinations of two unsigned 2-bit inputs.
a = pyrtl.Input(bitwidth=2, name="a")
b = pyrtl.Input(bitwidth=2, name="b")

unsigned_sum = pyrtl.Output(bitwidth=3, name="unsigned_sum")
unsigned_sum <<= a + b

# Try all combinations of {0, 1, 2, 3} + {0, 1, 2, 3}.
unsigned_inputs = {
    "a": [0, 0, 0, 0, 1, 1, 1, 1, 2, 2, 2, 2, 3, 3, 3, 3],
    "b": [0, 1, 2, 3, 0, 1, 2, 3, 0, 1, 2, 3, 0, 1, 2, 3],
}

sim = pyrtl.Simulation()
sim.step_multiple(provided_inputs=unsigned_inputs)

# In this trace, `unsigned_sum` is the sum of all combinations of
# {0, 1, 2, 3} + {0, 1, 2, 3}. For example:
# * cycle  0 shows 0 + 0 = 0
# * cycle  1 shows 0 + 1 = 1
# * cycle 15 shows 3 + 3 = 6
print(
    "Unsigned addition. Each cycle adds a different combination of numbers.\n"
    "unsigned_sum == a + b"
)
sim.tracer.render_trace(repr_func=int)

# ## Re-interpreting `a`, `b`, and `unsigned_sum` as signed is **incorrect**.
#
# Use `val_to_signed_integer()` to re-interpret the previous simulation results
# as signed integers. But the results are **INCORRECT**, because `unsigned_sum`
# performed unsigned addition with `+`. For example:
#
# * cycle  2 shows 0 + -2 = 2
# * cycle 13 shows -1 + 1 = -4
#
# `unsigned_sum` is incorrect because PyRTL must extend `a` and `b` to the
# sum's bitwidth before adding, but PyRTL zero-extends by default, instead of
# sign-extending. Zero-extending is correct when `a` and `b` are unsigned, but
# sign-extending is correct when `a` and `b` are signed. Use `signed_add()` to
# sign-extend `a` and `b`, as demonstrated next.
print(
    "\nUse `val_to_signed_integer()` to re-interpret the previous simulation results as"
    "\nsigned integers. But re-interpreting the previous trace as signed integers \n"
    "produces INCORRECT RESULTS!"
)
sim.tracer.render_trace(repr_func=pyrtl.val_to_signed_integer)

# ## Use `signed_add()` to correctly add signed integers.
#
# `signed_add()` works by sign-extending its inputs to the sum's bitwidth before
# adding. There are many `signed_*` functions for signed operations, like
# `signed_sub()`, `signed_mult()`, `signed_lt()`, and so on.
pyrtl.reset_working_block()

a = pyrtl.Input(bitwidth=2, name="a")
b = pyrtl.Input(bitwidth=2, name="b")

signed_sum = pyrtl.Output(bitwidth=3, name="signed_sum")
signed_sum <<= pyrtl.signed_add(a, b)

# Try all combinations of {-2, -1, 0, 1} + {-2, -1, 0, 1}.
signed_inputs = {
    "a": [-2, -2, -2, -2, -1, -1, -1, -1, 0, 0, 0, 0, 1, 1, 1, 1],
    "b": [-2, -1, 0, 1, -2, -1, 0, 1, -2, -1, 0, 1, -2, -1, 0, 1],
}
sim = pyrtl.Simulation()
sim.step_multiple(provided_inputs=signed_inputs)

# In this trace, `signed_sum` is the sum of all combinations of
# {-2, -1, 0, 1} + {-2, -1, 0, 1}. For example:
# * cycle  0 shows -2 + -2 = -4
# * cycle  1 shows -2 + -1 = -3
# * cycle 15 shows 1 + 1 = 2
print(
    "\nReset the simulation and use `signed_add()` to correctly add signed integers.\n"
    "signed_sum == signed_add(a, b)"
)
sim.tracer.render_trace(repr_func=pyrtl.val_to_signed_integer)

# ## Manually sign-extend inputs to correctly add signed integers with `+`.
#
# Instead of using `signed_add()`, we can manually sign-extend the inputs with
# `sign_extended()` and `truncate()` the output to correctly add signed integers with
# `+`. Use this trick to implement other signed arithmetic operations.
pyrtl.reset_working_block()

a = pyrtl.Input(bitwidth=2, name="a")
b = pyrtl.Input(bitwidth=2, name="b")

# Using `+` produces the correct result for signed addition here because we
# manually extend `a` and `b` to 3 bits. The result of
#
#     a.sign_extended(3) + b.sign_extended(3)
#
# is now 4 bits, but we truncate it to 3 bits. This truncation is subtle, but important!
# The addition's full 4 bit result would not be correct.
sign_extended_sum = pyrtl.Output(bitwidth=3, name="sign_extended_sum")
extended_a = a.sign_extended(bitwidth=3)
extended_b = b.sign_extended(bitwidth=3)
sign_extended_sum <<= (extended_a + extended_b).truncate(bitwidth=3)

# Try all combinations of {-2, -1, 0, 1} + {-2, -1, 0, 1}.
signed_inputs = {
    "a": [-2, -2, -2, -2, -1, -1, -1, -1, 0, 0, 0, 0, 1, 1, 1, 1],
    "b": [-2, -1, 0, 1, -2, -1, 0, 1, -2, -1, 0, 1, -2, -1, 0, 1],
}
sim = pyrtl.Simulation()
sim.step_multiple(provided_inputs=signed_inputs)

print(
    "\nInstead of using `signed_add()`, we can also manually sign extend the inputs to"
    "\ncorrectly add signed integers with `+`.\nsign_extended_sum == "
    "(a.sign_extended(3) + b.sign_extended(3)).truncate(3)"
)
sim.tracer.render_trace(repr_func=pyrtl.val_to_signed_integer)



import enum
import io
import random
import unittest

import pyrtl

graphviz_string_detailed = """\
digraph g {
    graph [splines="spline", outputorder="edgesfirst"];
    node [shape=circle, style=filled, fillcolor=lightblue1,
        fontcolor=black, fontname=helvetica, penwidth=0,
        fixedsize=shape];
    edge [labelfloat=false, penwidth=2, color=deepskyblue, arrowsize=.5];
    n0 [label="a", shape=invhouse, fillcolor=coral];
    n1 [label="8", shape=circle, fillcolor=lightgrey];
    n2 [label="0", shape=circle, fillcolor=lightgrey];
    n3 [label="0", shape=circle, fillcolor=lightgrey];
    n4 [label=" (Fanout: 0)", height=.1, width=.1];
    n5 [label="d", shape=house, fillcolor=lawngreen];
    n6 [label="[0]*2 (Fanout: 1)", fillcolor=azure1, height=.25, width=.25];
    n7 [label="concat (Fanout: 1)", height=.1, width=.1];
    n8 [label="* (Fanout: 1)"];
    n9 [label="[7:2] (Fanout: 1)", fillcolor=azure1, height=.25, width=.25];
    n10 [label="[0]*4 (Fanout: 1)", fillcolor=azure1, height=.25, width=.25];
    n11 [label="concat (Fanout: 1)", height=.1, width=.1];
    n0 -> n7 [label="a/2 (Delay: 0.00)", penwidth="6", arrowhead="none"];
    n1 -> n8 [label="const_0_8/4 (Delay: 0.00)", penwidth="6", arrowhead="normal"];
    n2 -> n6 [label="const_1_0/1 (Delay: 0.00)", penwidth="2", arrowhead="none"];
    n3 -> n10 [label="const_2_0/1 (Delay: 0.00)", penwidth="2", arrowhead="none"];
    n4 -> n5 [label="d/10 (Delay: 706.50)", penwidth="6", arrowhead="normal"];
    n6 -> n7 [label="tmp0/2 (Delay: 0.00)", penwidth="6", arrowhead="none"];
    n7 -> n8 [label="tmp1/4 (Delay: 0.00)", penwidth="6", arrowhead="normal"];
    n8 -> n9 [label="tmp2/8 (Delay: 706.50)", penwidth="6", arrowhead="none"];
    n9 -> n11 [label="tmp3/6 (Delay: 706.50)", penwidth="6", arrowhead="none"];
    n10 -> n11 [label="tmp4/4 (Delay: 0.00)", penwidth="6", arrowhead="none"];
    n11 -> n4 [label="tmp5/10 (Delay: 706.50)", penwidth="6", arrowhead="normal"];
    {
        rank=same;
        edge[style=invis];
        n6 -> n0;
        rankdir=LR;
    }
    {
        rank=same;
        edge[style=invis];
        n10 -> n9;
        rankdir=LR;
    }
}

"""

graphviz_string_arg_ordered = """\
digraph g {
    graph [splines="spline", outputorder="edgesfirst"];
    node [shape=circle, style=filled, fillcolor=lightblue1,
        fontcolor=black, fontname=helvetica, penwidth=0,
        fixedsize=shape];
    edge [labelfloat=false, penwidth=2, color=deepskyblue, arrowsize=.5];
    n0 [label="0", shape=circle, fillcolor=lightgrey];
    n1 [label="0", shape=circle, fillcolor=lightgrey];
    n2 [label="0", shape=circle, fillcolor=lightgrey];
    n3 [label="i", shape=invhouse, fillcolor=coral];
    n4 [label="j", shape=invhouse, fillcolor=coral];
    n5 [label="", height=.1, width=.1];
    n6 [label="o", shape=house, fillcolor=lawngreen];
    n7 [label="", height=.1, width=.1];
    n8 [label="q", shape=house, fillcolor=lawngreen];
    n9 [label="[0]*4", fillcolor=azure1, height=.25, width=.25];
    n10 [label="concat", height=.1, width=.1];
    n11 [label="<"];
    n12 [label="[0]*7", fillcolor=azure1, height=.25, width=.25];
    n13 [label="concat", height=.1, width=.1];
    n14 [label="[0]*4", fillcolor=azure1, height=.25, width=.25];
    n15 [label="concat", height=.1, width=.1];
    n16 [label=">"];
    n0 -> n9 [label="", penwidth="2", arrowhead="none"];
    n1 -> n12 [label="", penwidth="2", arrowhead="none"];
    n2 -> n14 [label="", penwidth="2", arrowhead="none"];
    n3 -> n11 [label="", penwidth="6", arrowhead="normal"];
    n3 -> n16 [label="", penwidth="6", arrowhead="normal"];
    n4 -> n10 [label="", penwidth="6", arrowhead="none"];
    n4 -> n15 [label="", penwidth="6", arrowhead="none"];
    n5 -> n6 [label="", penwidth="6", arrowhead="normal"];
    n7 -> n8 [label="", penwidth="2", arrowhead="normal"];
    n9 -> n10 [label="", penwidth="6", arrowhead="none"];
    n10 -> n11 [label="", penwidth="6", arrowhead="normal"];
    n11 -> n13 [label="", penwidth="2", arrowhead="none"];
    n12 -> n13 [label="", penwidth="6", arrowhead="none"];
    n13 -> n5 [label="", penwidth="6", arrowhead="normal"];
    n14 -> n15 [label="", penwidth="6", arrowhead="none"];
    n15 -> n16 [label="", penwidth="6", arrowhead="normal"];
    n16 -> n7 [label="", penwidth="2", arrowhead="normal"];
    {
        rank=same;
        edge[style=invis];
        n9 -> n4;
        rankdir=LR;
    }
    {
        rank=same;
        edge[style=invis];
        n3 -> n10;
        rankdir=LR;
    }
    {
        rank=same;
        edge[style=invis];
        n12 -> n11;
        rankdir=LR;
    }
    {
        rank=same;
        edge[style=invis];
        n14 -> n4;
        rankdir=LR;
    }
    {
        rank=same;
        edge[style=invis];
        n15 -> n3;
        rankdir=LR;
    }
}

"""

graphviz_string_arg_unordered = """\
digraph g {
    graph [splines="spline", outputorder="edgesfirst"];
    node [shape=circle, style=filled, fillcolor=lightblue1,
        fontcolor=black, fontname=helvetica, penwidth=0,
        fixedsize=shape];
    edge [labelfloat=false, penwidth=2, color=deepskyblue, arrowsize=.5];
    n0 [label="0", shape=circle, fillcolor=lightgrey];
    n1 [label="0", shape=circle, fillcolor=lightgrey];
    n2 [label="0", shape=circle, fillcolor=lightgrey];
    n3 [label="i", shape=invhouse, fillcolor=coral];
    n4 [label="j", shape=invhouse, fillcolor=coral];
    n5 [label="", height=.1, width=.1];
    n6 [label="o", shape=house, fillcolor=lawngreen];
    n7 [label="", height=.1, width=.1];
    n8 [label="q", shape=house, fillcolor=lawngreen];
    n9 [label="[0]*4", fillcolor=azure1, height=.25, width=.25];
    n10 [label="concat", height=.1, width=.1];
    n11 [label="<"];
    n12 [label="[0]*7", fillcolor=azure1, height=.25, width=.25];
    n13 [label="concat", height=.1, width=.1];
    n14 [label="[0]*4", fillcolor=azure1, height=.25, width=.25];
    n15 [label="concat", height=.1, width=.1];
    n16 [label=">"];
    n0 -> n9 [label="", penwidth="2", arrowhead="none"];
    n1 -> n12 [label="", penwidth="2", arrowhead="none"];
    n2 -> n14 [label="", penwidth="2", arrowhead="none"];
    n3 -> n11 [label="", penwidth="6", arrowhead="normal"];
    n3 -> n16 [label="", penwidth="6", arrowhead="normal"];
    n4 -> n10 [label="", penwidth="6", arrowhead="none"];
    n4 -> n15 [label="", penwidth="6", arrowhead="none"];
    n5 -> n6 [label="", penwidth="6", arrowhead="normal"];
    n7 -> n8 [label="", penwidth="2", arrowhead="normal"];
    n9 -> n10 [label="", penwidth="6", arrowhead="none"];
    n10 -> n11 [label="", penwidth="6", arrowhead="normal"];
    n11 -> n13 [label="", penwidth="2", arrowhead="none"];
    n12 -> n13 [label="", penwidth="6", arrowhead="none"];
    n13 -> n5 [label="", penwidth="6", arrowhead="normal"];
    n14 -> n15 [label="", penwidth="6", arrowhead="none"];
    n15 -> n16 [label="", penwidth="6", arrowhead="normal"];
    n16 -> n7 [label="", penwidth="2", arrowhead="normal"];
}

"""


class TestOutputGraphs(unittest.TestCase):
    def setUp(self):
        pyrtl.reset_working_block()

    def test_output_to_tgf_does_not_throw_error(self):
        from tests.test_importexport import full_adder_blif

        with io.StringIO() as vfile:
            pyrtl.input_from_blif(full_adder_blif)
            pyrtl.output_to_trivialgraph(vfile)

    def test_output_to_graphviz_does_not_throw_error(self):
        from tests.test_importexport import full_adder_blif

        with io.StringIO() as vfile:
            pyrtl.input_from_blif(full_adder_blif)
            pyrtl.output_to_graphviz(vfile)

    def test_output_to_graphviz_with_custom_namer_does_not_throw_error(self):
        from tests.test_importexport import full_adder_blif

        with io.StringIO() as vfile:
            pyrtl.input_from_blif(full_adder_blif)
            timing = pyrtl.TimingAnalysis()
            node_fan_in = {net: len(net.args) for net in pyrtl.working_block()}
            graph_namer = pyrtl.graphviz_detailed_namer(
                extra_node_info=node_fan_in, extra_edge_info=timing.timing_map
            )
            pyrtl.output_to_graphviz(vfile, namer=graph_namer)

    def test_output_to_graphviz_correct_detailed_output(self):
        pyrtl.wire._reset_wire_indexers()

        a = pyrtl.Input(2, "a")
        b = a * 8
        c = b[2:]
        d = pyrtl.Output(10, "d")
        d <<= c

        analysis = pyrtl.TimingAnalysis()
        _, dst_map = pyrtl.working_block().net_connections()

        def get_fanout(n):
            if isinstance(n, pyrtl.LogicNet):
                if n.op == "@":
                    return 0
                w = n.dests[0]
            else:
                w = n

            if isinstance(w, pyrtl.Output):
                return 0
            return len(dst_map[w])

        node_fanout = {
            n: f"Fanout: {get_fanout(n)}" for n in pyrtl.working_block().logic
        }
        wire_delay = {
            w: f"Delay: {analysis.timing_map[w]:.2f}"
            for w in pyrtl.working_block().wirevector_set
        }

        with io.StringIO() as vfile:
            pyrtl.output_to_graphviz(
                file=vfile,
                namer=pyrtl.graphviz_detailed_namer(node_fanout, wire_delay),
                maintain_arg_order=True,
            )
            self.assertEqual(vfile.getvalue(), graphviz_string_detailed)

    def test_output_to_graphviz_correct_output_with_arg_ordering(self):
        i = pyrtl.Input(8, "i")
        j = pyrtl.Input(4, "j")
        o = pyrtl.Output(8, "o")
        q = pyrtl.Output(1, "q")
        o <<= i < j
        q <<= j > i

        with io.StringIO() as vfile:
            pyrtl.output_to_graphviz(file=vfile, maintain_arg_order=True)
            self.assertEqual(vfile.getvalue(), graphviz_string_arg_ordered)

    def test_output_to_graphviz_correct_output_without_arg_ordering(self):
        i = pyrtl.Input(8, "i")
        j = pyrtl.Input(4, "j")
        o = pyrtl.Output(8, "o")
        q = pyrtl.Output(1, "q")
        o <<= i < j
        q <<= j > i

        with io.StringIO() as vfile:
            pyrtl.output_to_graphviz(file=vfile)
            self.assertEqual(vfile.getvalue(), graphviz_string_arg_unordered)


class TestNetGraph(unittest.TestCase):
    def setUp(self):
        pyrtl.reset_working_block()

    def test_as_graph(self):
        inwire = pyrtl.Input(bitwidth=1, name="inwire1")
        inwire2 = pyrtl.Input(bitwidth=1)
        inwire3 = pyrtl.Input(bitwidth=1)
        tempwire = pyrtl.WireVector()
        tempwire2 = pyrtl.WireVector()
        outwire = pyrtl.Output()

        tempwire <<= inwire | inwire2
        tempwire2 <<= ~tempwire
        outwire <<= tempwire2 & inwire3

        g = pyrtl.net_graph()
        # note for future: this might fail if we change the way that temp wires are
        # inserted, but that should not matter for this test and so the number can be
        # safely updated.
        self.assertEqual(len(g), 10)

        self.assertEqual(len(g[inwire]), 1)
        self.assertEqual(next(iter(g[inwire].keys())).op, "|")
        self.assertEqual(len(g[inwire].values()), 1)
        edges = next(iter(g[inwire].values()))
        self.assertEqual(len(edges), 1)
        self.assertIs(edges[0], inwire)

    def test_netgraph_unused_wires(self):
        _ = pyrtl.WireVector(8, "genwire")
        _ = pyrtl.Input(8, "inwire")
        _ = pyrtl.Output(8, "outwire")
        _ = pyrtl.Const(8, 8)
        _ = pyrtl.Register(8, "reg")

        g = pyrtl.net_graph()
        self.assertEqual(len(g), 0)

    def test_netgraph_same_wire_multiple_edges_to_same_net(self):
        c = pyrtl.Const(1, 1)
        w = pyrtl.concat(c, c, c)
        g = pyrtl.net_graph()
        self.assertEqual(len(g[c]), 1)
        edges = next(iter(g[c].values()))
        self.assertEqual(len(edges), 3)
        for w in edges:
            self.assertIs(w, c)


class TestOutputIPynb(unittest.TestCase):
    def setUp(self):
        pyrtl.reset_working_block()
        self.maxDiff = None

    def test_one_bit_adder_matches_expected(self):
        temp1 = pyrtl.WireVector(bitwidth=1, name="temp1")
        temp2 = pyrtl.WireVector()

        a, b, c = pyrtl.Input(1, "a"), pyrtl.Input(1, "b"), pyrtl.Input(1, "c")
        sum, carry_out = pyrtl.Output(1, "sum"), pyrtl.Output(1, "carry_out")

        sum <<= a ^ b ^ c

        temp1 <<= a & b  # connect the result of a & b to the pre-allocated wirevector
        temp2 <<= a & c
        temp3 = (
            b & c
        )  # temp3 IS the result of b & c (this is the first mention of temp3)
        carry_out <<= temp1 | temp2 | temp3

        sim = pyrtl.Simulation()
        for _cycle in range(15):
            sim.step(
                {
                    "a": random.choice([0, 1]),
                    "b": random.choice([0, 1]),
                    "c": random.choice([0, 1]),
                }
            )

        _ = pyrtl.trace_to_html(sim.tracer)  # tests if it compiles or not

    def test_trace_to_html(self):
        i = pyrtl.Input(1, "i")
        o = pyrtl.Output(2, "o")
        o <<= i + 1

        sim = pyrtl.Simulation()
        sim.step_multiple({"i": "0100110"})
        htmlstring = pyrtl.trace_to_html(sim.tracer)
        expected = (
            '<script type="WaveDrom">\n'
            "{\n"
            "  signal : [\n"
            '    { name: "i",  wave: "010.1.0" },\n'
            '    { name: "o",  wave: "===.=.=", data: ["0x1", "0x2", "0x1", "0x2", "0x1"] },\n'  # noqa: E501
            "  ],\n"
            "  config: { hscale: 1 }\n"
            "}\n"
            "</script>\n"
        )
        self.assertEqual(htmlstring, expected)

    def test_trace_to_html_repr_func(self):
        i = pyrtl.Input(1, "i")
        o = pyrtl.Output(2, "o")
        o <<= i + 1

        sim = pyrtl.Simulation()
        sim.step_multiple({"i": "0100110"})
        htmlstring = pyrtl.trace_to_html(sim.tracer, repr_func=bin)
        expected = (
            '<script type="WaveDrom">\n'
            "{\n"
            "  signal : [\n"
            '    { name: "i",  wave: "010.1.0" },\n'
            '    { name: "o",  wave: "===.=.=", data: ["0b1", "0b10", "0b1", "0b10", "0b1"] },\n'  # noqa: E501
            "  ],\n"
            "  config: { hscale: 1 }\n"
            "}\n"
            "</script>\n"
        )
        self.assertEqual(htmlstring, expected)

    def test_trace_to_html_repr_per_name(self):
        class Foo(enum.IntEnum):
            A = 0
            B = 1
            C = 2
            D = 3

        i = pyrtl.Input(4, "i")
        state = pyrtl.Register(max(Foo).bit_length(), name="state")
        o = pyrtl.Output(name="o")
        o <<= state

        with pyrtl.conditional_assignment:
            with i == 0b0001:
                state.next |= Foo.A
            with i == 0b0010:
                state.next |= Foo.B
            with i == 0b0100:
                state.next |= Foo.C
            with i == 0b1000:
                state.next |= Foo.D

        sim = pyrtl.Simulation()
        sim.step_multiple({"i": [1, 2, 4, 8, 0]})

        htmlstring = pyrtl.trace_to_html(
            sim.tracer, repr_per_name={"state": pyrtl.enum_name(Foo)}
        )
        expected = (
            '<script type="WaveDrom">\n'
            "{\n"
            "  signal : [\n"
            '    { name: "i",  wave: "=====", data: ["0x1", "0x2", "0x4", "0x8", "0x0"] },\n'  # noqa: E501
            '    { name: "o",  wave: "=.===", data: ["0x0", "0x1", "0x2", "0x3"] },\n'
            '    { name: "state",  wave: "=.===", data: ["A", "B", "C", "D"] },\n'
            "  ],\n"
            "  config: { hscale: 1 }\n"
            "}\n"
            "</script>\n"
        )
        self.assertEqual(htmlstring, expected)

    def test_trace_to_html_repr_per_name_enum_is_bool(self):
        class Foo(enum.IntEnum):
            A = 0
            B = 1

        i = pyrtl.Input(2, "i")
        state = pyrtl.Register(max(Foo).bit_length(), name="state")
        o = pyrtl.Output(name="o")
        o <<= state

        with pyrtl.conditional_assignment:
            with i == 0b01:
                state.next |= Foo.A
            with i == 0b10:
                state.next |= Foo.B

        sim = pyrtl.Simulation()
        sim.step_multiple({"i": [1, 2, 1, 2, 2]})

        htmlstring = pyrtl.trace_to_html(
            sim.tracer, repr_per_name={"state": pyrtl.enum_name(Foo)}
        )
        expected = (
            '<script type="WaveDrom">\n'
            "{\n"
            "  signal : [\n"
            '    { name: "i",  wave: "====.", data: ["0x1", "0x2", "0x1", "0x2"] },\n'
            '    { name: "o",  wave: "0.101" },\n'
            '    { name: "state",  wave: "=.===", data: ["A", "B", "A", "B"] },\n'
            "  ],\n"
            "  config: { hscale: 1 }\n"
            "}\n"
            "</script>\n"
        )
        self.assertEqual(htmlstring, expected)


if __name__ == "__main__":
    unittest.main()



"""
Contains functions to estimate aspects of blocks (like area and delay) by either using
internal models or by making calls out to external tool chains.
"""

from __future__ import annotations

import collections
import math
import os
import re
import subprocess
import sys
import tempfile
from collections.abc import Iterable
from typing import Callable

from pyrtl.core import Block, LogicNet, working_block
from pyrtl.helperfuncs import _currently_in_jupyter_notebook, _print_netlist_latex
from pyrtl.importexport import output_to_verilog
from pyrtl.memory import RomBlock
from pyrtl.pyrtlexceptions import PyrtlError, PyrtlInternalError
from pyrtl.wire import Const, Input, Output, Register, WireVector

# --------------------------------------------------------------------
#         __   ___          ___  __  ___              ___    __
#    /\  |__) |__   /\     |__  /__`  |  |  |\/|  /\   |  | /  \ |\ |
#   /~~\ |  \ |___ /~~\    |___ .__/  |  |  |  | /~~\  |  | \__/ | \|
#


def area_estimation(tech_in_nm: float = 130, block=None) -> tuple[float, float]:
    """Estimates the total area of the block.

    The estimations are based off of 130nm standard cell designs for the logic, and
    custom memory blocks from the literature. The results are not fully validated and we
    do not recommend that this function be used in carrying out science for publication.

    :param tech_in_nm: the size of the circuit technology to be estimated (for example,
        65 is 65nm and 250 is 0.25um)

    :return: tuple of estimated areas (logic, mem) in terms of mm^2
    """

    def mem_area_estimate(tech_in_nm, bits, ports, is_rom):
        # http://www.cs.ucsb.edu/~sherwood/pubs/ICCD-srammodel.pdf
        # ROM is assumed to be 1/10th of area of SRAM
        tech_in_um = tech_in_nm / 1000.0
        area_estimate = 0.001 * tech_in_um**2.07 * bits**0.9 * ports**0.7 + 0.0048
        return area_estimate if not is_rom else area_estimate / 10.0

    # Subset of the raw data gathered from yosys, mapping to vsclib 130nm library
    # Width   Adder_Area  Mult_Area  (area in "tracks" as discussed below)
    # 8       211         2684
    # 16      495         12742
    # 32      1110        49319
    # 64      2397        199175
    # 128     4966        749828

    def adder_stdcell_estimate(width):
        return width * 34.4 - 25.8

    def multiplier_stdcell_estimate(width):
        if width == 1:
            return 5
        if width == 2:
            return 39
        if width == 3:
            return 219
        return -958 + (150 * width) + (45 * width**2)

    def stdcell_estimate(net):
        if net.op in "w~sc":
            return 0
        if net.op in "&|n":
            return 40 / 8.0 * len(net.args[0])  # 40 lambda
        if net.op in "^=<>x":
            return 80 / 8.0 * len(net.args[0])  # 80 lambda
        if net.op == "r":
            return 144 / 8.0 * len(net.args[0])  # 144 lambda
        if net.op in "+-":
            return adder_stdcell_estimate(len(net.args[0]))
        if net.op == "*":
            return multiplier_stdcell_estimate(len(net.args[0]))
        if net.op in "m@":
            return 0  # memories handled elsewhere
        msg = (
            "Unable to estimate the following net due to unimplemented op :\n{str(net)}"
        )
        raise PyrtlInternalError(msg)

    block = working_block(block)

    # The functions above were gathered and calibrated by mapping reference designs to
    # an openly available 130nm stdcell library.
    # http://www.vlsitechnology.org/html/vsc_description.html
    # http://www.vlsitechnology.org/html/cells/vsclib013/lib_gif_index.html

    # In a standard cell design, each gate takes up a length of standard "track" in the
    # chip. The functions above return that length for each of the different types of
    # functions in the units of "tracks". In the 130nm process used, 1 lambda is 55nm,
    # and 1 track is 8 lambda.

    # first, sum up the area of all of the logic elements (including registers)
    total_tracks = sum(stdcell_estimate(a_net) for a_net in block.logic)
    total_length_in_nm = total_tracks * 8 * 55
    # each track is then 72 lambda tall, and converted from nm2 to mm2
    area_in_mm2_for_130nm = (total_length_in_nm * (72 * 55)) / 1e12

    # scaling from 130nm to the target tech
    logic_area = area_in_mm2_for_130nm / (130.0 / tech_in_nm) ** 2

    # now sum up the area of the memories
    mem_area = 0
    for mem in {net.op_param[1] for net in block.logic_subset("@m")}:
        bits, ports, is_rom = _bits_ports_and_isrom_from_memory(mem)
        mem_area += mem_area_estimate(tech_in_nm, bits, ports, is_rom)

    return logic_area, mem_area


def _bits_ports_and_isrom_from_memory(mem):
    """Helper to extract mem bits and ports for estimation."""
    is_rom = False
    bits = 2**mem.addrwidth * mem.bitwidth
    read_ports = len(mem.readport_nets)
    write_ports = len(mem.writeport_nets)
    if isinstance(mem, RomBlock):
        is_rom = True
    ports = max(read_ports, write_ports)
    return bits, ports, is_rom


# --------------------------------------------------------------------
#   ___                 __        /\                     __      __
#    |  |  |\/| | |\ | /  `      /~~\ |\ |  /\  |  \_/  /__` |  /__`
#    |  |  |  | | | \| \__>     /    \| \| /~~\ |_  |   .__/ |  .__/
#


class TimingAnalysis:
    """Timing analysis estimates the timing delays in the block

    TimingAnalysis has an :attr:`timing_map` object that maps wires to the 'time' after
    a clock edge at which the signal in the wire settles
    """

    def __init__(self, block: Block = None, gate_delay_funcs=None):
        """Calculates timing delays in the block.

        Calculates the timing analysis while allowing for different timing delays of
        different gates of each type. Supports all valid presynthesis blocks. Currently
        doesn't support memory post synthesis.

        :param block: PyRTL block to analyze. Defaults to the :ref:`working_block`.
        :param gate_delay_funcs: a map with keys corresponding to the gate op and a
            function returning the delay as the value. It takes the gate as an argument.
            If the delay is negative (``-1``), the gate will be treated as the end of
            the block.
        """

        self.block = working_block(block)
        self.timing_map = None
        self.block.sanity_check()
        self._generate_timing_map(gate_delay_funcs)

    def _generate_timing_map(self, gate_delay_funcs):
        # The functions above were gathered and calibrated by mapping reference designs
        # to an openly available 130nm stdcell library. Note that this is will compute
        # the critical logic delay, but does not include setup/hold time.

        if gate_delay_funcs is None:
            gate_delay_funcs = {
                "~": lambda _width: 48.5,
                "&": lambda _width: 98.5,
                "|": lambda _width: 105.3,
                "^": lambda _width: 135.07,
                "n": lambda _width: 66.0,
                "w": lambda _width: 0,
                "+": self._logconst_func(184.0, 18.9),
                "-": self._logconst_func(184.0, 18.9),
                "*": self._multiplier_stdcell_estimate,
                "<": self._logconst_func(101.9, 105.4),
                ">": self._logconst_func(101.9, 105.4),
                "=": self._logconst_func(60.1, 147),
                "x": lambda _width: 138.0,
                "c": lambda _width: 0,
                "s": lambda _width: 0,
                "r": lambda _width: -1,
                "m": self._memory_read_estimate,
                "@": lambda _width: -1,
            }
        cleared = self.block.wirevector_subset((Input, Const, Register))
        self.timing_map = dict.fromkeys(cleared, 0)
        for _gate in self.block:  # ordered iteration
            if _gate.op == "m":
                gate_delay = gate_delay_funcs["m"](
                    _gate.op_param[1]
                )  # reads require a memid
            else:
                gate_delay = gate_delay_funcs[_gate.op](len(_gate.args[0]))

            if gate_delay < 0:
                continue
            time = max(self.timing_map[a_wire] for a_wire in _gate.args) + gate_delay
            for dest_wire in _gate.dests:
                self.timing_map[dest_wire] = time

    @staticmethod
    def _logconst_func(a, b):
        return lambda x: a * math.log2(float(x)) + b

    @staticmethod
    def _multiplier_stdcell_estimate(width):
        if width == 1:
            return 98.57
        if width == 2:
            return 200.17
        return 549.1 * math.log2(width) - 391.7

    @staticmethod
    def _memory_read_estimate(mem):
        # http://www.cs.ucsb.edu/~sherwood/pubs/ICCD-srammodel.pdf
        # ROM is assumed to be same delay as SRAM (perhaps optimistic?)
        bits, ports, is_rom = _bits_ports_and_isrom_from_memory(mem)
        tech_in_um = 0.130
        return 270 * tech_in_um**1.38 * bits**0.25 * ports**1.30 + 1.05

    def max_freq(
        self, tech_in_nm: float = 130, ffoverhead: float | None = None
    ) -> float:
        """Estimates the max frequency of a block in MHz.

        All params are optional and have reasonable default values. Estimation is based
        on Dennard Scaling assumption and does not include wiring effect -- as a result
        the estimates may be optimistic (especially below 65nm).

        :param tech_in_nm: the size of the circuit technology to be estimated (for
            example, 65 is 65nm and 250 is 0.25um)
        :param ffoverhead: setup and ff propagation delay in picoseconds

        :return: a number representing an estimate of the max frequency in Mhz
        """
        cp_length = self.max_length()
        scale_factor = 130.0 / tech_in_nm
        if ffoverhead is None:
            clock_period_in_ps = scale_factor * (cp_length + 189 + 194)
        else:
            clock_period_in_ps = (scale_factor * cp_length) + ffoverhead
        return 1e6 * 1.0 / clock_period_in_ps

    def max_length(self):
        """Returns the max timing delay of the circuit in ps.

        The result assumes that the circuit is implemented in a 130nm process, and that
        there is no setup or hold time associated with the circuit. The resulting value
        is in picoseconds. If an proper estimation of timing is required it is
        recommended to use :meth:`max_freq` to determine the clock period as it more
        accurately considers scaling and setup/hold.
        """
        return max(self.timing_map.values())

    def print_max_length(self):
        """Prints the max timing delay of the circuit"""
        print("The total block timing delay is ", self.max_length())

    class _TooManyCPsError(Exception):
        pass

    def critical_path(
        self, print_cp: bool = True, cp_limit=100
    ) -> list[WireVector, list[LogicNet]]:
        """Takes a timing map and returns the critical paths of the system.

        :param print_cp: Whether to print the critical path to the terminal after
            calculation

        :return: a list containing tuples with the 'first' wire as the first value and
                 the critical paths (which themselves are lists of nets) as the second
        """
        critical_paths = []  # storage of all completed critical paths
        wire_src_map, dst_map = self.block.net_connections()

        def critical_path_pass(old_critical_path, first_wire):
            if isinstance(first_wire, (Input, Const, Register)):
                critical_paths.append((first_wire, old_critical_path))
                return

            if len(critical_paths) >= cp_limit:
                raise self._TooManyCPsError()

            source = wire_src_map[first_wire]
            critical_path = [source]
            critical_path.extend(old_critical_path)
            arg_max_time = max(self.timing_map[arg_wire] for arg_wire in source.args)
            for arg_wire in source.args:
                # If the time for both items are the max, both will be on a critical
                # path.
                if self.timing_map[arg_wire] == arg_max_time:
                    critical_path_pass(critical_path, arg_wire)

        max_time = self.max_length()
        try:
            for wire_pair in self.timing_map.items():
                if wire_pair[1] == max_time:
                    critical_path_pass([], wire_pair[0])
        except self._TooManyCPsError:
            print("Critical path count limit reached")

        if print_cp:
            self.print_critical_paths(critical_paths)
        return critical_paths

    @staticmethod
    def print_critical_paths(critical_paths):
        """Prints the results of the critical path length analysis. Done by default by
        the :meth:`critical_path` function.
        """
        line_indent = " " * 2
        #  print the critical path
        for cp_with_num in enumerate(critical_paths):
            print("Critical path", cp_with_num[0], ":")
            print(line_indent, "The first wire is:", cp_with_num[1][0])
            if _currently_in_jupyter_notebook():
                _print_netlist_latex(cp_with_num[1][1])
            else:
                for net in cp_with_num[1][1]:
                    print(line_indent, (net))
            print()


# --------------------------------------------------------------------
#          __   __       __
#     \ / /  \ /__` \ / /__`
#      |  \__/ .__/  |  .__/
#


def yosys_area_delay(
    library: str,
    abc_cmd: str | None = None,
    leave_in_dir: str | None = None,
    block: Block = None,
) -> tuple[float, float]:
    """Synthesize with `Yosys <https://yosyshq.net/yosys/>`_ and return estimate of area
    and delay.

    If ``leave_in_dir`` is specified, that directory will be used to create any
    temporary files, and the resulting files will be left behind there (which can be
    useful for manual exploration or debugging)

    The area and delay are returned in units as defined by the stdcell library. In the
    standard vsc 130nm library, the area is in a number of "tracks", each of which is
    about 1.74 square um (see area estimation for more details) and the delay is in ps.

    http://www.vlsitechnology.org/html/vsc_description.html

    :param library: stdcell library file to target in liberty format
    :param abc_cmd: string of commands for :program:`yosys` to pass to :program:`abc`
        for synthesis
    :param leave_in_dir: the directory where temporary files should be left
    :param block: PyRTL block to analyze. Defaults to the :ref:`working_block`.

    :raise PyrtlError: If :program:`yosys` is not configured correctly.
    :raise PyrtlInternalError: If the call to :program:`yosys` was not successful

    :return: a tuple of numbers: area, delay
    """

    if abc_cmd is None:
        abc_cmd = "strash;scorr;ifraig;retime;dch,-f;map;print_stats;"
    else:
        # first, replace whitespace with commas as per yosys requirements
        re.sub(r"\s+", ",", abc_cmd)
        # then append with "print_stats" to generate the area and delay info
        abc_cmd = f"{abc_cmd};print_stats;"

    def extract_area_delay_from_yosys_output(yosys_output):
        report_lines = [
            line for line in yosys_output.decode().split("\n") if "ABC: netlist" in line
        ]
        area = re.match(r".*area\s*=\s*([0-9\.]*)", report_lines[0]).group(1)
        delay = re.match(r".*delay\s*=\s*([0-9\.]*)", report_lines[0]).group(1)
        return float(area), float(delay)

    yosys_arg_template = """-p
    read_verilog %s;
    synth -top toplevel;
    dfflibmap -liberty %s;
    abc -liberty %s -script +%s
    """.replace("\n", " ")

    temp_d, temp_path = tempfile.mkstemp(
        prefix="pyrtl_verilog", suffix=".v", dir=leave_in_dir, text=True
    )
    try:
        # write the verilog to a temp
        yosys_arg = yosys_arg_template % (temp_path, library, library, abc_cmd)
        with open(temp_path, "w") as f:
            print("// generated via pyrtl yosys_area_delay", file=f)
            print(f"// yosys {yosys_arg}", file=f)
            output_to_verilog(f, block=block)
        os.close(temp_d)
        # call yosys on the temp, and grab the output
        yosys_output = subprocess.check_output(["yosys", yosys_arg])
        area, delay = extract_area_delay_from_yosys_output(yosys_output)
    except (subprocess.CalledProcessError, ValueError) as exc:
        print("Error with call to yosys...", file=sys.stderr)
        print("---------------------------------------------", file=sys.stderr)
        print(str(exc.output).replace("\\n", "\n"), file=sys.stderr)
        print("---------------------------------------------", file=sys.stderr)
        msg = "Yosys call failed"
        raise PyrtlError(msg) from exc
    except OSError as exc:
        print("Error with call to yosys...", file=sys.stderr)
        msg = "Call to yosys failed (not installed or on path?)"
        raise PyrtlError(msg) from exc
    finally:
        if leave_in_dir is None:
            os.remove(temp_path)
    return area, delay


class PathsResult(dict):
    def print(self, file=sys.stdout):
        """Pretty print the result of calling :func:`paths`

        :param f: the open file to print to (defaults to stdout)

        :return: None
        """

        # All this work, to make sure it's determinstic
        def path_sort_key(path):
            dst_names = [net.dests[0].name if net.dests else "" for net in path]
            return (len(path), dst_names)

        for start in sorted(self.keys(), key=lambda w: w.name):
            print(f"From {start.name}", file=file)
            for end in sorted(self[start].keys(), key=lambda w: w.name):
                print(f"  To {end.name}", file=file)
                all_paths = self[start][end]
                if len(all_paths) > 0:
                    for i, paths in enumerate(sorted(all_paths, key=path_sort_key)):
                        print(f"    Path {i}", file=file)
                        for path in paths:
                            print(f"      {path}", file=file)
                else:
                    print("    (No paths)", file=file)


def paths(
    src: WireVector | Iterable[WireVector] = None,
    dst: WireVector | Iterable[WireVector] = None,
    dst_nets: dict[WireVector, LogicNet] | None = None,
    block: Block = None,
) -> PathsResult:
    """Get the list of all paths from ``src`` to ``dst``.

    You can provide ``dst_nets`` (the result of calling :meth:`Block.net_connections`,
    if you plan on calling this function repeatedly on a block that hasn't changed, to
    speed things up.

    This function can accept one or more ``src`` wires, and one or more ``dst`` wires,
    such that it returns a map that can be accessed like so::

        paths[src][dst] = [<path>, <path>, ...]

    where ``path`` is a list of nets. Thus there can be multiple paths from a given
    ``src`` wire to a given ``dst`` wire.

    If ``src`` and ``dst`` are both single wires, you still need to access the result
    via ``paths[src][dst]``.

    This also finds and returns the loop paths in the case of registers or memories that
    feed into themselves, i.e. ``paths[src][src]`` is not necessarily empty.

    It does not distinguish between loops that include synchronous vs asynchronous
    memories.

    :param src: Source wire(s) from which to trace your paths; if ``None``, will get
        paths from all :class:`Inputs<Input>`.
    :param dst: Destination wire(s) to which to trace your paths; if ``None``, will get
        paths to all :class:`Outputs<Output>`.
    :param dst_nets: map from wire to set of nets where the wire is an argument; will
        compute it internally if not given via a call to
        :meth:`Block.net_connections()`.
    :param block: Block to use (defaults to :ref:`working_block`)

    :return: a map of the form ``{src_wire: {dst_wire: [path]}}`` for each ``src_wire``
             in ``src`` (or all inputs if ``src`` is None), ``dst_wire`` in ``dst`` (or
             all outputs if ``dst`` is None), where ``path`` is a list of nets. This map
             is also an instance of :class:`.PathsResult`, so you can call
             :meth:`.PathsResult.print` on it to pretty print it.
    """
    block = working_block(block)

    if dst_nets is None:
        # Note: if you set `include_virtual_nodes=True`, Output wires will actually be
        # present as the destination "net" of Output wires in the dst_nets map. That
        # would overly complicate this algorithm: we will assume all values() in the
        # dst_nets map are logic nets only. We set this to False for explicitness...
        _, dst_nets = block.net_connections(include_virtual_nodes=False)
    else:
        # ... or make sure it's not present otherwise.
        for output in block.wirevector_subset(cls=Output):
            dst_nets.pop(output, None)

    if src is None:
        src = block.wirevector_subset(cls=Input)
    elif isinstance(src, WireVector):
        src = {src}
    else:
        src = set(src)

    if dst is None:
        dst = block.wirevector_subset(cls=Output)
    elif isinstance(dst, WireVector):
        dst = {dst}
    else:
        dst = set(dst)

    def paths_src_dst(src, dst):
        paths = []

        # Use DFS to get the paths [each a list of nets] from src wire to dst wire
        def dfs(w, curr_path):
            if w is dst:
                # Found valid path
                paths.append(curr_path)
            for dst_net in dst_nets.get(w, []):
                # Avoid loops and the mem net (has no output wire)
                if dst_net not in curr_path:
                    if dst_net.op == "@":  # dests will be the read ports
                        for read_net in dst_net.op_param[1].readport_nets:
                            dfs(read_net.dests[0], [*curr_path, dst_net, read_net])
                    else:
                        dfs(dst_net.dests[0], [*curr_path, dst_net])

        dfs(src, [])
        return paths

    all_paths = collections.defaultdict(dict)
    for src_wire in src:
        for dst_wire in dst:
            paths = paths_src_dst(src_wire, dst_wire)
            # Remove empty paths...
            paths = list(filter(lambda x: len(x) > 0, paths))
            # ...and those that are supersets of others (resulting from an inner loop).
            if src_wire is not dst_wire:
                paths = sorted(paths, key=lambda p: len(p), reverse=True)
                keep = []
                for i in range(len(paths)):
                    # Check if there is a path in paths[i+1:] that is the suffix of
                    # paths[i] (paths[i] is at least as large as each path in
                    # paths[i+1:]). If so, paths[i] contains a loop since both start at
                    # src_wire, so don't keep it.
                    if not any(paths[i][-len(p) :] == p for p in paths[i + 1 :]):
                        keep.append(paths[i])
                paths = keep
            all_paths[src_wire][dst_wire] = paths

    return PathsResult(all_paths)


def distance(
    src: WireVector,
    dst: WireVector,
    f: Callable[list[LogicNet], int],
    block: Block = None,
) -> dict[tuple[LogicNet], int]:
    """Calculate the distance along each path from ``src`` to ``dst`` according to ``f``

    This calls the given function ``f`` on each net in a path, summing the result.

    :param src: wire to start from
    :param dst: wire to end on
    :param f: function from a net to number, representing the 'value' of a net that you
              want to sum across all nets in the path
    :param block: block to use (defaults to :ref:`working_block`)

    :return: a map from each path (a tuple) to its calculated distance
    """
    ps = paths(src, dst, block=block)
    ps = ps[src][dst]
    # Turning path into tuple so it can be the key
    return {tuple(path): sum(map(f, path)) for path in ps}


def fanout(w: WireVector) -> int:
    """Get the number of places a wire is used as an argument.

    :param w: :class:`WireVector` to check fanout for.

    :return: Integer fanout count.
    """
    _, dst_nets = w._block.net_connections()
    if w not in dst_nets:
        return 0

    all_args = [arg for net in dst_nets[w] for arg in net.args]
    return len(list(filter(lambda arg: arg is w, all_args)))



# # Example 3: A State Machine built with `conditional_assignment`
#
# In this example we describe how `conditional_assignment` works in the context of a
# vending machine that will dispense an item when it has received 4 tokens. If a refund
# is requested, it returns the tokens.
import enum

import pyrtl

# Define `Inputs`, `Outputs`, and a `Register` to keep track of the vending machine's
# current state.
token_in = pyrtl.Input(1, "token_in")
req_refund = pyrtl.Input(1, "req_refund")

dispense = pyrtl.Output(1, "dispense")
refund = pyrtl.Output(1, "refund")

state = pyrtl.Register(3, "state")


# First new step, let's enumerate a set of constants to serve as our states
class State(enum.IntEnum):
    WAIT = 0  # Waiting for first token.
    TOK1 = 1  # Received first token, waiting for second token.
    TOK2 = 2  # Received second token, waiting for third token.
    TOK3 = 3  # Received third token, waiting for fourth token.
    DISP = 4  # Received fourth token, dispense item.
    RFND = 5  # Issue refund.


# Now we could build a state machine using just the `Registers` and logic discussed in
# prior examples, but doing operations **conditionally** on some input is a pretty
# fundamental operation in hardware design. PyRTL provides `conditional_assignment` to
# provide a predicated update to `Registers`, `WireVectors`, and `MemBlocks`.
#
# `conditional_assignments` are specified with the `|=` operator instead of the usual
# `<<=` operator. The `conditional_assignment` is only valid in the context of a
# condition, and updates to those values only happens when that condition is `True`. In
# hardware this is implemented with a simple `mux()` -- for people coming from software
# it is important to remember that this is describing a big logic function, **NOT** an
# "if-then-else" clause. The `conditional_assignment` below just builds hardware
# multiplexers. Nothing conditional actually happens until the `Simulation` is run.
#
# One more thing: `conditional_assignment` might not always be the best solution. For
# simple updates, a regular
#
#     select(sel_wire, truecase=t_wire, falsecase=f_wire)
#
# can be easier to read.
with pyrtl.conditional_assignment:
    with req_refund:  # signal of highest precedence
        state.next |= State.RFND
    with token_in:  # if token received, advance state in counter sequence
        with state == State.WAIT:
            state.next |= State.TOK1
        with state == State.TOK1:
            state.next |= State.TOK2
        with state == State.TOK2:
            state.next |= State.TOK3
        with state == State.TOK3:
            state.next |= State.DISP  # 4th token received, go to dispense
        with pyrtl.otherwise:  # token received in unsupported state
            state.next |= State.RFND
    # unconditional transition from these two states back to wait state

    # NOTE: the parens are needed because in Python the "|" operator is lower
    # precedence than the "==" operator!
    with (state == State.DISP) | (state == State.RFND):
        state.next |= State.WAIT

dispense <<= state == State.DISP
refund <<= state == State.RFND

# A few more notes:
#
# 1. A condition can be nested within another condition and the implied hardware is that
#    the left-hand-side should only get that value if ALL of the encompassing conditions
#    are satisfied.
# 2. Only one conditional at each level can be `True`, so all conditions imply that NONE
#    of the prior conditions at the same level are `True`. The highest priority
#    condition is listed first, and in a sense you can think about each other condition
#    as an `elif`.
# 3. If a `WireVector`'s value is not specified for some combination of conditions,
#    `conditional_assignment` will supply a default value. By default,`Registers` will
#    retain their value from the prior cycle ("state.next |= state" in this example).
#    `WireVectors` default to `0`.
# 4. There is a way to specify something like an `else` instead of `elif` and that is
#    with an `otherwise` (as seen on the line above "state.next <<= State.RFND"). This
#    condition will be `True` if none of the other conditions at the same level were
#    also `True`. For this example specifically, `state.next` will get `RFND` when
#    `req_refund==0`, and `token_in==1`, and state is not any of {`WAIT`, `TOK1`,
#    `TOK2`, or `TOK3`}.
# 5. Not shown here, but you can update multiple `Registers`, `WireVectors`, and
#    `MemBlocks` within one `conditional_assignment`.
#
# A more artificial example might make it even more clear how these rules interact:
#
#     with a:
#         r.next |= 1        # ← When `a` is `True`.
#         with d:
#             r2.next |= 2   # ← When `a` is `True` and `d` is `True`.
#         with pyrtl.otherwise:
#             r2.next |= 3   # ← When `a` is `True` and `d` is `False`.
#     with b == c:
#         r.next |= 0        # ← When `a` is `False` and `b == c`.
#
# Now let's build and test our state machine.
sim = pyrtl.Simulation()

# Rather than just give some random inputs, let's specify some specific 1-bit values. To
# make it easier to `Simulate` over several steps, we'll use `step_multiple()`, which
# takes in a `dict` mapping each `Input` to its value in each cycle.
sim_inputs = {"token_in": "0010100111010000", "req_refund": "1100010000000000"}
sim.step_multiple(sim_inputs)

# Also, to make our input/output easy to reason about let's specify an order to the
# traces with `trace_list`. We also use `enum_name` to display the state names (`WAIT`,
# `TOK1`, ...) rather than their numbers (0, 1, ...).
sim.tracer.render_trace(
    trace_list=["token_in", "req_refund", "state", "dispense", "refund"],
    repr_per_name={"state": pyrtl.enum_name(State)},
)

# Finally, suppose you want to simulate your design and verify its output matches your
# expectations. `step_multiple()` also accepts as a second argument a `dict` mapping
# output wires to their expected value in each cycle. If during the `Simulation` the
# actual and expected values differ, it will be reported to you! This might be useful if
# you have a working design which, after some tweaks, you'd like to test for functional
# equivalence, or as a basic sanity check.
expected_sim_outputs = {"dispense": "0000000000001000", "refund": "0111001000000000"}
sim = pyrtl.Simulation()
sim.step_multiple(sim_inputs, expected_sim_outputs)
